import binascii
import sys
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, check_and_set_session


class IOCBI(UDSScanner):
    """Input output control"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session",
            type=auto_int,
            default=0x01,
            help="The session in which the requests are made",
        )
        self.parser.add_argument(
            "data_identifier", type=auto_int, help="The data identifier"
        )
        self.parser.add_argument(
            "control_parameter",
            type=str,
            choices=[
                "return-control-to-ecu",
                "reset-to-default",
                "freeze-current-state",
                "short-term-adjustment",
                "without-control-parameter",
            ],
            help='Control parameter sent to the ECU. "short-term-adjustment" and "without-control-parameter"'
            " require passing a new state as well.",
        )
        self.parser.add_argument(
            "--new-state",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help='The new state required in use with the two control parameters "short-term-adjustment"'
            ' and "without-control-parameter".',
        )
        self.parser.add_argument(
            "--control-enable-mask",
            metavar="HEXSTRING",
            type=binascii.unhexlify,
            default=b"",
            help="This parameter is used if the data-identifier corresponds to multiple signals."
            "In that case each bit enables or disables setting of each corresponding signal."
            "Can only be used in combination with a control parameter.",
        )

    async def main(self, args: Namespace) -> None:
        try:
            await check_and_set_session(self.ecu, args.session)
        except Exception as e:
            self.logger.log_critical(
                f"Could not change to session: 0x{args.session:02x}: {e.__class__.__name__} {e}"
            )
            sys.exit(1)

        did = args.data_identifier
        control_enable_mask_record = args.control_enable_mask
        uses_control_parameter = True

        if args.control_parameter == "return-control-to-ecu":
            resp = (
                await self.ecu.input_output_control_by_identifier_return_control_to_ecu(
                    did, control_enable_mask_record
                )
            )
        elif args.control_parameter == "reset-to-default":
            resp = await self.ecu.input_output_control_by_identifier_reset_to_default(
                did, control_enable_mask_record
            )
        elif args.control_parameter == "freeze-current-state":
            resp = (
                await self.ecu.input_output_control_by_identifier_freeze_current_state(
                    did, control_enable_mask_record
                )
            )
        elif args.control_parameter == "short-term-adjustment":
            resp = (
                await self.ecu.input_output_control_by_identifier_short_term_adjustment(
                    did, args.new_state, control_enable_mask_record
                )
            )
        elif args.control_parameter == "without-control-parameter":
            resp = await self.ecu.input_output_control_by_identifier(
                did, args.new_state, control_enable_mask_record
            )
            uses_control_parameter = False
        else:
            self.logger.log_critical("Unhandled control parameter")
            sys.exit(1)

        if isinstance(resp, NegativeResponse):
            self.logger.log_error(resp)
        else:
            self.logger.log_summary("Positive response:")
            data = (
                resp.control_status_record[1:]
                if uses_control_parameter
                else resp.control_status_record
            )
            self.logger.log_summary(f"hex: {data.hex()}")
            self.logger.log_summary(f"raw: {repr(data)}")



import asyncio
import binascii
import reprlib
from argparse import Namespace
from itertools import product

from gallia.uds.core.constants import UDSErrorCodes, RCSubFuncs, UDSIsoServices
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.core.exception import IllegalResponse
from gallia.uds.core.service import NegativeResponse, UDSResponse
from gallia.uds.helpers import suggests_service_not_supported
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import (
    auto_int,
    find_sessions,
    check_and_set_session,
    ParseSkips,
)


class ScanIdentifiers(UDSScanner):
    """This scanner scans DataIdentifiers of various
    services. Specific requirements such as for RoutineControl or SecurityAccess
    are considered and implemented in the script.
    """

    def add_parser(self) -> None:
        self.parser.add_argument(
            "--sessions",
            type=auto_int,
            nargs="*",
            help="Set list of sessions to be tested; all if None",
        )
        self.parser.add_argument(
            "--start",
            type=auto_int,
            default=0,
            help="start scan at this dataIdentifier (default: 0x%(default)x)",
        )
        self.parser.add_argument(
            "--end",
            type=auto_int,
            default=0xFFFF,
            help="end scan at this dataIdentifier (default: 0x%(default)x)",
        )
        self.parser.add_argument(
            "--payload",
            default=None,
            type=binascii.unhexlify,
            help="Payload which will be appended for each request as hex string",
        )
        self.parser.add_argument(
            "--sid",
            type=auto_int,
            default=0x22,
            help="""
            Service ID to scan; defaults to ReadDataByIdentifier (default: 0x%(default)x);
            currently supported:
            0x27 Security Access;
            0x22 Read Data By Identifier;
            0x2e Write Data By Identifier;
            0x31 Routine Control;
            """,
        )
        self.parser.add_argument(
            "--check-session",
            nargs="?",
            const=1,
            type=int,
            help="Check current session via read DID [for every nth DataIdentifier] and try to recover session",
        )
        self.parser.add_argument(
            "--skip",
            nargs="+",
            default={},
            type=str,
            action=ParseSkips,
            help="""
                 The data identifiers to be skipped per session.
                 A session specific skip is given by <session_id>:<identifiers>
                 where <identifiers> is a comma separated list of single ids or id ranges using a dash.
                 Examples:
                  - 0x01:0xf3
                  - 0x10-0x2f
                  - 0x01:0xf3,0x10-0x2f
                 Multiple session specific skips are separated by space.
                 """,
        )

    async def main(self, args: Namespace) -> None:
        if args.sessions is None:
            self.logger.log_info("No sessions specified, starting with session scan")
            # Only until 0x80 because the eight bit is "SuppressResponse"
            sessions = list(
                s
                for s in range(1, 0x80)
                if s not in args.skip or args.skip[s] is not None
            )
            sessions = await find_sessions(self.ecu, sessions)
            self.logger.log_summary(
                f'Found {len(sessions)} sessions: {" ".join([hex(i) for i in sessions])}'
            )
        else:
            sessions = list(
                s
                for s in args.sessions
                if s not in args.skip or args.skip[s] is not None
            )

        self.logger.log_info(f"testing sessions {sessions}")
        self.logger.log_info(f"skipping identifiers {reprlib.repr(args.skip)}")

        for session in sessions:
            self.logger.log_notice(f"Switching to session 0x{session:02x}")
            resp: UDSResponse = await self.ecu.set_session(session)
            if isinstance(resp, NegativeResponse):
                self.logger.log_warning(
                    f"Switching to session 0x{session:02x} failed: {resp}"
                )
                continue

            self.logger.log_summary(f"Starting scan in session: 0x{session:02x}")
            positive_DIDs = 0
            abnormal_DIDs = 0
            timeout_DIDs = 0
            sub_functions = [0x00]

            if args.sid == UDSIsoServices.RoutineControl:
                if not args.payload:
                    self.logger.log_warning(
                        "Scanning RoutineControl with empty payload might start some routines "
                        "and change the state of the ECU!"
                    )

                # Scan all three subfunctions (startRoutine, stopRoutine, requestRoutineResults)
                sub_functions = list(map(int, RCSubFuncs))

            if args.sid == UDSIsoServices.SecurityAccess:
                if args.end > 0xFF:
                    self.logger.log_warning(
                        "Service 0x27 SecurityAccess only accepts 1-byte identifiers; "
                        + f"limiting END to 0xff instead of 0x{args.end:0x}"
                    )
                    args.end = 0xFF

            for (DID, sub_function) in product(
                range(args.start, args.end + 1), sub_functions
            ):
                if session in args.skip and DID in args.skip[session]:
                    self.logger.log_info(f"0x{DID:0x}: skipped")
                    continue

                if args.check_session and DID % args.check_session == 0:
                    # Check session and try to recover from wrong session (max 3 times), else skip session
                    if not await check_and_set_session(self.ecu, session):
                        self.logger.log_error(
                            f"Aborting scan on session 0x{session:02x}; current DID was 0x{DID:0x}"
                        )
                        break

                if args.sid == UDSIsoServices.SecurityAccess:
                    if DID & 0b10000000:
                        self.logger.log_info(
                            "Keep in mind that you set the SuppressResponse Bit (8th bit): "
                            + f"0x{DID:0x} = 0b{DID:b}"
                        )
                    pdu = bytes([args.sid, DID])

                elif args.sid == UDSIsoServices.RoutineControl:
                    pdu = bytes(
                        [args.sid, sub_function, DID >> 8, DID & 0xFF]
                    )  # Needs extra byte for sub function

                # DefaultBehaviour, e.g. for ReadDataByIdentifier/WriteDataByIdentifier
                else:
                    pdu = bytes([args.sid, DID >> 8, DID & 0xFF])

                if args.payload:
                    pdu += args.payload

                try:
                    resp = await self.ecu.send_raw(
                        pdu, config=UDSRequestConfig(tags=["ANALYZE"], max_retry=3)
                    )
                except asyncio.TimeoutError:
                    self.logger.log_summary(f"0x{DID:0x}: Retries exceeded")
                    timeout_DIDs += 1
                    continue
                except IllegalResponse as e:
                    self.logger.log_warning(f"{repr(e)}")

                if isinstance(resp, NegativeResponse):
                    if suggests_service_not_supported(resp):
                        self.logger.log_info(
                            f"0x{DID:0x}: {resp}; does session 0x{session:02x} "
                            f"support service 0x{args.sid:02x}?"
                        )

                    # RequestOutOfRange is a common reply for invalid DataIdentifiers
                    elif resp.response_code == UDSErrorCodes.requestOutOfRange:
                        self.logger.log_info(f"0x{DID:0x}: {resp}")

                    else:
                        self.logger.log_summary(f"0x{DID:0x}: {resp}")
                        abnormal_DIDs += 1
                else:
                    self.logger.log_summary(f"0x{DID:0x}: {resp}")
                    positive_DIDs += 1

            self.logger.log_summary(f"Scan in session 0x{session:0x} is complete!")
            self.logger.log_summary(f"Positive replies: {positive_DIDs}")
            self.logger.log_summary(f"Abnormal replies: {abnormal_DIDs}")
            self.logger.log_summary(f"Timeouts: {timeout_DIDs}")

            self.logger.log_info(f"Leaving session 0x{session:02x} via hook")
            await self.ecu.leave_session(session)



from __future__ import annotations

import asyncio
import errno
import time
import socket as s
import struct
from typing import cast, Optional, TypedDict

from can import Message

from gallia.transports.base import BaseTransport, _int_spec, _bool_spec, TargetURI


CANFD_MTU = 72
CAN_MTU = 16


_ISOTP_SPEC_TYPE = TypedDict(
    "_ISOTP_SPEC_TYPE",
    {
        "src_addr": int,
        "dst_addr": int,
        "is_extended": bool,
        "is_fd": bool,
        "frame_txtime": int,
        "ext_address": Optional[int],
        "rx_ext_address": Optional[int],
        "tx_padding": Optional[int],
        "rx_padding": Optional[int],
        "tx_dl": Optional[int],
    },
)

isotp_spec = {
    "src_addr": (_int_spec(0), True),
    "dst_addr": (_int_spec(0), True),
    "is_extended": (_bool_spec(False), False),
    "is_fd": (_bool_spec(False), False),
    "frame_txtime": (_int_spec(10), False),
    "ext_address": (_int_spec(None), False),
    "rx_ext_address": (_int_spec(None), False),
    "tx_padding": (_int_spec(None), False),
    "rx_padding": (_int_spec(None), False),
    "tx_dl": (_int_spec(None), False),
}


class ISOTPTransport(BaseTransport, scheme="isotp", spec=isotp_spec):
    # Socket Constants not available in the socket module,
    # see linux/can/isotp.h
    # TODO: Can be removed in the future…
    # https://github.com/python/cpython/pull/23794
    SOL_CAN_ISOTP = s.SOL_CAN_BASE + s.CAN_ISOTP

    # Valuetypes for SOL_CAN_ISOTP
    CAN_ISOTP_OPTS = 1
    CAN_ISOTP_RECV_FC = 2
    CAN_ISOTP_TX_STMIN = 3
    CAN_ISOTP_RX_STMIN = 4
    CAN_ISOTP_LL_OPTS = 5

    # Flags for setsockopt CAN_ISOTP_OPTS
    CAN_ISOTP_LISTEN_MODE = 0x001
    CAN_ISOTP_EXTEND_ADDR = 0x002
    CAN_ISOTP_TX_PADDING = 0x004
    CAN_ISOTP_RX_PADDING = 0x008
    CAN_ISOTP_CHK_PAD_LEN = 0x010
    CAN_ISOTP_CHK_PAD_DATA = 0x020
    CAN_ISOTP_HALF_DUPLEX = 0x040
    CAN_ISOTP_FORCE_TXSTMIN = 0x080
    CAN_ISOTP_FORCE_RXSTMIN = 0x100
    CAN_ISOTP_RX_EXT_ADDR = 0x200

    def __init__(self, target: TargetURI) -> None:
        super().__init__(target)
        self.args = cast(_ISOTP_SPEC_TYPE, self._args)

        assert target.hostname is not None, "empty interface"
        self.interface = target.hostname
        self._sock: s.socket

    async def connect(self, timeout: Optional[float] = None) -> None:
        self._sock = s.socket(s.PF_CAN, s.SOCK_DGRAM, s.CAN_ISOTP)
        self._sock.setblocking(False)

        src_addr = self._set_flags(self.args["src_addr"], self.args["is_extended"])
        dst_addr = self._set_flags(self.args["dst_addr"], self.args["is_extended"])

        self._setsockopts(
            frame_txtime=self.args["frame_txtime"],
            ext_address=self.args["ext_address"],
            rx_ext_address=self.args["rx_ext_address"],
            tx_padding=self.args["tx_padding"],
            rx_padding=self.args["rx_padding"],
        )
        # If CAN-FD is used, jumbo frames are possible.
        # This fails for non-fd configurations.
        if self.args["is_fd"]:
            tx_dl = 64 if self.args["tx_dl"] is None else self.args["tx_dl"]
            self._setsockllopts(canfd=self.args["is_fd"], tx_dl=tx_dl)
        self._sock.bind((self.interface, dst_addr, src_addr))

    @staticmethod
    def _set_flags(can_id: int, extended: bool = False) -> int:
        if extended:
            can_id = can_id & s.CAN_EFF_MASK
            return can_id | s.CAN_EFF_FLAG
        return can_id & s.CAN_SFF_MASK

    def _setsockopts(
        self,
        frame_txtime: int,
        tx_padding: Optional[int] = None,
        rx_padding: Optional[int] = None,
        ext_address: Optional[int] = None,
        rx_ext_address: Optional[int] = None,
        flags: int = 0,
    ) -> None:
        if ext_address:
            flags |= self.CAN_ISOTP_EXTEND_ADDR
        else:
            ext_address = 0

        if rx_ext_address:
            flags |= self.CAN_ISOTP_RX_EXT_ADDR
        else:
            rx_ext_address = 0

        if tx_padding:
            flags |= self.CAN_ISOTP_TX_PADDING
        else:
            tx_padding = 0

        if rx_padding:
            flags |= self.CAN_ISOTP_RX_PADDING
        else:
            rx_padding = 0

        data = struct.pack(
            "@IIBBBB",
            flags,
            frame_txtime,
            ext_address,
            tx_padding,
            rx_padding,
            rx_ext_address,
        )
        self._sock.setsockopt(self.SOL_CAN_ISOTP, self.CAN_ISOTP_OPTS, data)

    def _setsockfcopts(self, bs: int = 0, stmin: int = 0, wftmax: int = 0) -> None:
        data = struct.pack("@BBB", bs, stmin, wftmax)
        self._sock.setsockopt(self.SOL_CAN_ISOTP, self.CAN_ISOTP_RECV_FC, data)

    def _setsockllopts(self, canfd: bool, tx_dl: int) -> None:
        canmtu = 72 if canfd else 16
        # The flags are set to 0, since the author marks this as obsolete.
        data = struct.pack("@BBB", canmtu, tx_dl, 0)
        self._sock.setsockopt(self.SOL_CAN_ISOTP, self.CAN_ISOTP_LL_OPTS, data)

    async def write(
        self,
        data: bytes,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        self.logger.log_write(data.hex(), tags=tags)
        loop = asyncio.get_running_loop()
        await asyncio.wait_for(loop.sock_sendall(self._sock, data), timeout)
        return len(data)

    async def read(
        self, timeout: Optional[float] = None, tags: Optional[list[str]] = None
    ) -> bytes:
        loop = asyncio.get_running_loop()
        try:
            data = await asyncio.wait_for(
                loop.sock_recv(self._sock, self.BUFSIZE), timeout
            )
        except OSError as e:
            if e.errno == errno.ECOMM:
                raise BrokenPipeError(f"isotp flow control frame missing: {e}") from e
            if e.errno == errno.EILSEQ:
                raise BrokenPipeError(f"invalid consecutive frame numbers: {e}") from e
            raise e
        self.logger.log_read(data.hex(), tags=tags)
        return data

    async def sendto(
        self,
        data: bytes,
        dst: int,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        raise NotImplementedError()

    async def recvfrom(
        self, timeout: Optional[float] = None, tags: Optional[list[str]] = None
    ) -> tuple[int, bytes]:
        raise NotImplementedError()

    async def terminate(self) -> None:
        pass

    async def reconnect(self, timeout: Optional[float] = None) -> None:
        pass


class CANMessage(Message):  # type: ignore

    # https://git.kernel.org/pub/scm/linux/kernel/git/torvalds/linux.git/tree/include/uapi/linux/can.h
    CAN_HEADER_FMT = struct.Struct("=IBB2x")
    CANFD_BRS = 0x01
    CANFD_ESI = 0x02

    def _compose_arbitration_id(self) -> int:
        can_id = self.arbitration_id
        if self.is_extended_id:
            can_id |= s.CAN_EFF_FLAG
        if self.is_remote_frame:
            can_id |= s.CAN_RTR_FLAG
        if self.is_error_frame:
            can_id |= s.CAN_ERR_FLAG
        return can_id

    def pack(self) -> bytes:
        can_id = self._compose_arbitration_id()
        flags = 0
        if self.bitrate_switch:
            flags |= self.CANFD_BRS
        if self.error_state_indicator:
            flags |= self.CANFD_ESI
        max_len = 64 if self.is_fd else 8
        data = bytes(self.data).ljust(max_len, b"\x00")
        return self.CAN_HEADER_FMT.pack(can_id, self.dlc, flags) + data

    @staticmethod
    def _dissect_can_frame(frame: bytes) -> tuple[int, int, int, bytes]:
        can_id, can_dlc, flags = CANMessage.CAN_HEADER_FMT.unpack_from(frame)
        if len(frame) != CANFD_MTU:
            # Flags not valid in non-FD frames
            flags = 0
        return can_id, can_dlc, flags, frame[8 : 8 + can_dlc]

    @classmethod
    def unpack(cls, frame: bytes) -> CANMessage:
        can_id, can_dlc, flags, data = cls._dissect_can_frame(frame)

        # EXT, RTR, ERR flags -> boolean attributes
        #   /* special address description flags for the CAN_ID */
        #   #define CAN_EFF_FLAG 0x80000000U /* EFF/SFF is set in the MSB */
        #   #define CAN_RTR_FLAG 0x40000000U /* remote transmission request */
        #   #define CAN_ERR_FLAG 0x20000000U /* error frame */
        is_extended_frame_format = bool(can_id & s.CAN_EFF_FLAG)
        is_remote_transmission_request = bool(can_id & s.CAN_RTR_FLAG)
        is_error_frame = bool(can_id & s.CAN_ERR_FLAG)
        is_fd = len(frame) == CANFD_MTU
        bitrate_switch = bool(flags & cls.CANFD_BRS)
        error_state_indicator = bool(flags & cls.CANFD_ESI)

        if is_extended_frame_format:
            arbitration_id = can_id & s.CAN_EFF_MASK
        else:
            arbitration_id = can_id & s.CAN_SFF_MASK

        return cls(
            arbitration_id=arbitration_id,
            is_extended_id=is_extended_frame_format,
            is_remote_frame=is_remote_transmission_request,
            is_error_frame=is_error_frame,
            is_fd=is_fd,
            bitrate_switch=bitrate_switch,
            error_state_indicator=error_state_indicator,
            dlc=can_dlc,
            data=data,
        )


_CAN_RAW_SPEC_TYPE = TypedDict(
    "_CAN_RAW_SPEC_TYPE",
    {
        "src_addr": Optional[int],
        "dst_addr": Optional[int],
        "is_extended": bool,
        "is_fd": bool,
        "bind": bool,
    },
)

spec_can_raw = {
    "src_addr": (_int_spec(None), False),
    "dst_addr": (_int_spec(None), False),
    "is_extended": (_bool_spec(False), False),
    "is_fd": (_bool_spec(False), False),
    "bind": (_bool_spec(False), False),
}


class RawCANTransport(BaseTransport, scheme="can-raw", spec=spec_can_raw):
    # Flags for setsockopt CAN_RAW_FILTER
    CAN_INV_FILTER = 0x20000000

    def __init__(self, target: TargetURI) -> None:
        super().__init__(target)
        self.args = cast(_CAN_RAW_SPEC_TYPE, self._args)
        self.connected = False

        assert target.hostname is not None, "empty interface"
        self.interface = target.hostname
        self._sock: s.socket
        self.src_addr: int
        self.dst_addr: int

    async def connect(self, timeout: Optional[float] = None) -> None:
        self._sock = s.socket(s.PF_CAN, s.SOCK_RAW, s.CAN_RAW)
        self._sock.bind((self.interface,))
        if self.args["is_fd"] is True:
            self._sock.setsockopt(s.SOL_CAN_RAW, s.CAN_RAW_FD_FRAMES, 1)
        self._sock.setblocking(False)

        if self.args["bind"]:
            self.bind()

    def set_filter(self, can_ids: list[int], inv_filter: bool = False) -> None:
        if not can_ids:
            return
        filter_mask = s.CAN_EFF_MASK if self.args["is_extended"] else s.CAN_SFF_MASK
        data = b""
        for can_id in can_ids:
            if inv_filter:
                can_id |= self.CAN_INV_FILTER
            data += struct.pack("@II", can_id, filter_mask)
        self._sock.setsockopt(s.SOL_CAN_RAW, s.CAN_RAW_FILTER, data)
        if inv_filter:
            self._sock.setsockopt(s.SOL_CAN_RAW, s.CAN_RAW_JOIN_FILTERS, 1)

    def bind(self) -> None:
        if self.args["src_addr"] is None or self.args["dst_addr"] is None:
            raise RuntimeError("no src_addr/dst_addr set")

        self.set_filter([self.args["src_addr"]])
        self.src_addr = self.args["src_addr"]
        self.dst_addr = self.args["dst_addr"]
        self.connected = True

    async def read(
        self, timeout: Optional[float] = None, tags: Optional[list[str]] = None
    ) -> bytes:
        if not self.connected or not self.src_addr:
            raise RuntimeError("transport is not connected; set bind=true")
        _, data = await self.recvfrom(timeout, tags)
        return data

    async def write(
        self,
        data: bytes,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        if not self.connected or not self.dst_addr:
            raise RuntimeError("transport is not connected; set bind=true")
        return await self.sendto(data, self.dst_addr, timeout=timeout, tags=tags)

    async def sendto(
        self,
        data: bytes,
        dst: int,
        timeout: Optional[float] = None,
        tags: Optional[list[str]] = None,
    ) -> int:
        msg = CANMessage(
            arbitration_id=dst,
            data=data,
            is_extended_id=self.args["is_extended"],
            is_fd=self.args["is_fd"],
            check=True,
        )
        if self.args["is_extended"]:
            self.logger.log_write(f"{dst:08x}#{data.hex()}", tags=tags)
        else:
            self.logger.log_write(f"{dst:03x}#{data.hex()}", tags=tags)

        loop = asyncio.get_running_loop()
        await asyncio.wait_for(loop.sock_sendall(self._sock, msg.pack()), timeout)
        return len(data)

    async def recvfrom(
        self, timeout: Optional[float] = None, tags: Optional[list[str]] = None
    ) -> tuple[int, bytes]:
        loop = asyncio.get_running_loop()
        can_frame = await asyncio.wait_for(
            loop.sock_recv(self._sock, self.BUFSIZE), timeout
        )
        msg = CANMessage.unpack(can_frame)

        if msg.is_extended_id:
            self.logger.log_read(
                f"{msg.arbitration_id:08x}#{msg.data.hex()}", tags=tags
            )
        else:
            self.logger.log_read(
                f"{msg.arbitration_id:03x}#{msg.data.hex()}", tags=tags
            )
        return msg.arbitration_id, msg.data

    async def terminate(self) -> None:
        pass

    async def reconnect(self, timeout: Optional[float] = None) -> None:
        pass

    async def get_idle_traffic(self, sniff_time: float) -> list[int]:
        """Listen to traffic on the bus and return list of IDs
        which are seen in the specified period of time.
        The output of this function can be used as input to set_filter.
        """
        addr_idle: list[int] = list()
        t1 = time.time()
        while time.time() - t1 < sniff_time:
            try:
                addr, _ = await self.recvfrom(timeout=1)
                if addr not in addr_idle:
                    self.logger.log_info(f"Received a message from {addr:03x}")
                    addr_idle.append(addr)
            except asyncio.TimeoutError:
                continue
        addr_idle.sort()
        return addr_idle



from __future__ import annotations

import asyncio
import struct
from dataclasses import dataclass
from typing import Optional, overload, Sequence, Union

from gallia.uds.core import service
from gallia.uds.core.constants import UDSErrorCodes, UDSIsoServices

from gallia.uds.core.exception import MissingResponse
from gallia.uds.helpers import parse_pdu
from gallia.penlog import Logger
from gallia.transports.base import BaseTransport


@dataclass
class UDSRequestConfig:
    timeout: Optional[float] = None  # specify a timeout for this request
    max_retry: Optional[
        int
    ] = None  # maximum number of attempts in case of network errors
    skip_hooks: bool = False  # skip hooks
    tags: Optional[list[str]] = None  # tags to be applied to the logged output


class UDSClient:
    def __init__(
        self,
        transport: BaseTransport,
        timeout: Optional[float] = None,
        max_retry: int = 1,
    ):
        self.transport = transport
        self.timeout = timeout
        self.max_retry = max_retry
        self.retry_wait = 0.2
        self.pending_timeout = 5
        self.logger = Logger("uds", flush=True)

    async def reconnect(self, timeout: Optional[int] = None) -> None:
        """Calls the underlying transport to trigger a reconnect"""
        await self.transport.reconnect(timeout)

    async def _read(
        self, timeout: Optional[float] = None, tags: Optional[list[str]] = None
    ) -> bytes:
        if timeout is None and self.timeout:
            timeout = self.timeout
        return await self.transport.read(timeout, tags)

    async def request_unsafe(
        self, request: service.UDSRequest, config: Optional[UDSRequestConfig] = None
    ) -> service.UDSResponse:
        """This method is the same as request() with the difference
        that it does not hold the mutex in the underlying transport.
        """
        config = config if config is not None else UDSRequestConfig()

        last_exception: Exception = MissingResponse(request)
        max_retry = config.max_retry if config.max_retry else self.max_retry
        timeout = config.timeout if config.timeout else self.timeout
        for i in range(0, max_retry):
            # Exponential backoff
            wait_time = self.retry_wait * 2**i

            # Avoid pasting this very line in every error branch.
            if i > 0:
                self.logger.log_debug(f"retrying {i} from {max_retry}…")
            try:
                raw_resp = await self.transport.request_unsafe(
                    request.pdu, timeout, config.tags
                )
            except asyncio.TimeoutError as e:
                self.logger.log_debug(f"{request} failed with: {type(e)}")
                last_exception = MissingResponse(request, str(e))
                await asyncio.sleep(wait_time)
                continue

            resp = parse_pdu(raw_resp, request)

            if isinstance(resp, service.NegativeResponse):
                if resp.response_code == UDSErrorCodes.busyRepeatRequest:
                    if i >= max_retry - 1:
                        return resp
                    await asyncio.sleep(wait_time)
                    continue
            # We already had ECUs which thought an infinite
            # response_pending loop is a good idea…
            # Let's limit this.
            n_pending = 0
            n_timeout = 0
            waiting_time = 0.5
            max_n_timeout = max(timeout if timeout else 0, 10) / waiting_time
            while (
                isinstance(resp, service.NegativeResponse)
                and resp.response_code
                == UDSErrorCodes.requestCorrectlyReceivedResponsePending
            ):
                try:
                    raw_resp = await self._read(timeout=waiting_time, tags=config.tags)
                except asyncio.TimeoutError as e:
                    # Send a tester present to indicate that
                    # we are still there.
                    await self._tester_present(supress_resp=True)
                    n_timeout += 1
                    if n_timeout >= max_n_timeout:
                        last_exception = MissingResponse(request, str(e))
                        break
                    continue
                resp = parse_pdu(raw_resp, request)
                n_timeout = 0  # Only raise errors for consecutive timeouts
                n_pending += 1
                if n_pending >= 120:
                    raise RuntimeError(
                        "ECU appears to be stuck in ResponsePending loop"
                    )
            else:
                # We reach this code here once all response pending
                # and similar busy stuff is resolved.
                return resp

        self.logger.log_debug(f"{request} failed after retry loop")
        raise last_exception

    async def _tester_present(
        self, supress_resp: bool = False, config: Optional[UDSRequestConfig] = None
    ) -> Optional[service.UDSResponse]:
        config = config if config is not None else UDSRequestConfig()
        timeout = config.timeout if config.timeout else self.timeout
        if supress_resp:
            pdu = service.TesterPresentRequest(suppress_response=True).pdu
            await self.transport.write(pdu, timeout, config.tags)
            return None
        return await self.tester_present(False, config)

    async def send_raw(
        self, pdu: bytes, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.PositiveResponse]:
        """Raw request, which does not need to be compliant with the standard.
        It can be used to send arbitrary data packets.

        :param pdu: The data.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(service.RawRequest(pdu), config)

    async def diagnostic_session_control(
        self,
        diagnostic_session_type: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.DiagnosticSessionControlResponse]:
        """Sets the diagnostic session which is specified by a specific diagnosticSessionType
        sub-function.
        This is an implementation of the UDS request for service DiagnosticSessionControl (0x10).

        :param diagnostic_session_type: The session sub-function.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.DiagnosticSessionControlRequest(
                diagnostic_session_type, suppress_response
            ),
            config,
        )

    async def ecu_reset(
        self,
        reset_type: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ECUResetResponse]:
        """Resets the ECU using the specified reset type sub-function.
        This is an implementation of the UDS request for service ECUReset (0x11).

        :param reset_type: The reset type sub-function.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ECUResetRequest(reset_type, suppress_response), config
        )

    async def security_access_request_seed(
        self,
        security_access_type: int,
        security_access_data_record: bytes = b"",
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.SecurityAccessResponse]:
        """Requests a seed for a security access level.
        This is an implementation of the UDS request for the requestSeed sub-function group
        of the service SecurityAccess (0x27).

        :param security_access_type: The securityAccess type sub-function.
        :param security_access_data_record: Optional data.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.RequestSeedRequest(
                security_access_type, security_access_data_record, suppress_response
            ),
            config,
        )

    async def security_access_send_key(
        self,
        security_access_type: int,
        security_key: bytes,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.SecurityAccessResponse]:
        """Sends the key for a security access level.
        This is an implementation of the UDS request for the sendKey sub-function group
        of the service SecurityAccess (0x27).

        :param security_access_type: The securityAccess type sub-function.
        :param security_key: The response to the seed challenge.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.SendKeyRequest(
                security_access_type, security_key, suppress_response
            ),
            config,
        )

    async def communication_control(
        self,
        control_type: int,
        communication_type: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.CommunicationControlResponse]:
        """Controls communication of the ECU.
        This is an implementation of the UDS request for service CommunicationControl (0x28).

        :param control_type: The control type sub-function.
        :param communication_type: The communication type.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.CommunicationControlRequest(
                control_type, communication_type, suppress_response
            ),
            config,
        )

    async def tester_present(
        self, suppress_response: bool = False, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.TesterPresentResponse]:
        """Signals to the ECU, that the tester is still present.
        This is an implementation of the UDS request for service TesterPresent (0x3E).

        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.TesterPresentRequest(suppress_response), config
        )

    async def control_dtc_setting(
        self,
        dtc_setting_type: int,
        dtc_setting_control_option_record: bytes = b"",
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ControlDTCSettingResponse]:
        """Control the setting of DTCs.
        This is an implementation of the UDS request for service ControlDTCSetting (0x85).


        :param dtc_setting_type: The setting type.
        :param dtc_setting_control_option_record: Optional data.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ControlDTCSettingRequest(
                dtc_setting_type, dtc_setting_control_option_record, suppress_response
            ),
            config,
        )

    async def read_data_by_identifier(
        self,
        data_identifiers: Union[int, Sequence[int]],
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReadDataByIdentifierResponse]:
        """Reads data which is identified by a specific dataIdentifier.
        This is an implementation of the UDS request for service ReadDataByIdentifier (0x22).
        While this implementation supports requesting multiple dataIdentifiers at once, as is
        permitted in the standard, it is recommended to request them separately,
        because the support is optional on the server side.
        Additionally, it is not possible to reliably determine each single dataRecord from a
        corresponding response.

        :param data_identifiers: One or multiple dataIdentifiers. A dataIdentifier is a max two
                                 bytes integer.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReadDataByIdentifierRequest(data_identifiers), config
        )

    async def read_memory_by_address(
        self,
        memory_address: int,
        memory_size: int,
        address_and_length_format_identifier: Optional[int] = None,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReadMemoryByAddressResponse]:
        """Reads data from a specific memory address on the UDS server.
        This is an implementation of the UDS request for service ReadMemoryByAddress (0x3d).
        While it exposes each parameter of the corresponding specification,
        some parameters can be computed from the remaining ones and can therefore be omitted.

        :param memory_address: The start address.
        :param memory_size: The number of bytes to read.
        :param address_and_length_format_identifier: The byte lengths of the memory address and
                                                     size. If omitted, this parameter is computed
                                                     based on the memory_address and memory_size
                                                     parameters.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReadMemoryByAddressRequest(
                memory_address, memory_size, address_and_length_format_identifier
            ),
            config,
        )

    async def write_data_by_identifier(
        self,
        data_identifier: int,
        data_record: bytes,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.WriteDataByIdentifierResponse]:
        """Writes data which is identified by a specific dataIdentifier.
        This is an implementation of the UDS request for service WriteDataByIdentifier (0x2E).

        :param data_identifier: The identifier. A dataIdentifier is a max two bytes integer.
        :param data_record: The data to be written.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.WriteDataByIdentifierRequest(data_identifier, data_record), config
        )

    async def write_memory_by_address(
        self,
        memory_address: int,
        data_record: bytes,
        memory_size: Optional[int] = None,
        address_and_length_format_identifier: Optional[int] = None,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.WriteMemoryByAddressResponse]:
        """Writes data to a specific memory on the UDS server.
        This is an implementation of the UDS request for service writeMemoryByAddress (0x3d).
        While it exposes each parameter of the corresponding specification,
        some parameters can be computed from the remaining ones and can therefore be omitted.

        :param memory_address: The start address.
        :param data_record: The data to be written.
        :param memory_size: The number of bytes to write.
                            If omitted, the byte length of the data is used.
        :param address_and_length_format_identifier: The byte lengths of the memory address and
                                                     size. If omitted, this parameter is computed
                                                     based on the memory_address and memory_size
                                                     or data_record parameters.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.WriteMemoryByAddressRequest(
                memory_address,
                data_record,
                memory_size,
                address_and_length_format_identifier,
            ),
            config,
        )

    async def clear_diagnostic_information(
        self, group_of_dtc: int, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.ClearDiagnosticInformationResponse]:
        """Clears diagnostic trouble codes according to a given mask.
        This is an implementation of the UDS request for service clearDiagnosticInformation (0x14).

        :param group_of_dtc: The three byte mask, which determines the DTCs to be cleared.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ClearDiagnosticInformationRequest(group_of_dtc), config
        )

    async def read_dtc_information_report_number_of_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReportNumberOfDTCByStatusMaskResponse]:
        """Read the number of DTCs with the specified state from the UDS server.
        This is an implementation of the UDS request for the reportNumberOfDTCByStatusMask
        sub-function of the service ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportNumberOfDTCByStatusMaskRequest(
                dtc_status_mask, suppress_response
            ),
            config,
        )

    async def read_dtc_information_report_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReportDTCByStatusMaskResponse]:
        """Read DTCs and their state from the UDS server.
        This is an implementation of the UDS request for the reportDTCByStatusMask sub-function of
        the service ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
         response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportDTCByStatusMaskRequest(dtc_status_mask, suppress_response),
            config,
        )

    async def read_dtc_information_report_mirror_memory_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.ReportMirrorMemoryDTCByStatusMaskResponse
    ]:
        """Read DTCs and their state from the UDS server's mirror memory.
        This is an implementation of the UDS request for the reportMirrorMemoryDTCByStatusMask
        sub-function of the
        service ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportMirrorMemoryDTCByStatusMaskRequest(
                dtc_status_mask, suppress_response
            ),
            config,
        )

    async def read_dtc_information_report_number_of_mirror_memory_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportNumberOfMirrorMemoryDTCByStatusMaskResponse,
    ]:
        """Read the number of DTCs with the specified state from the UDS server's mirror memory.
        This is an implementation of the UDS request for the
        reportNumberOfMirrorMemoryDTCByStatusMask sub-function of
        the service ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportNumberOfMirrorMemoryDTCByStatusMaskRequest(
                dtc_status_mask, suppress_response
            ),
            config,
        )

    async def read_dtc_information_report_number_of_emissions_related_obd_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportNumberOfEmissionsRelatedOBDDTCByStatusMaskResponse,
    ]:
        """Read the number of emission related DTCs with the specified state from the UDS server.
        This is an implementation of the UDS request for the
        reportNumberOfEmissionsRelatedOBDDTCByStatusMask sub-function of the service
        ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportNumberOfEmissionsRelatedOBDDTCByStatusMaskRequest(
                dtc_status_mask, suppress_response
            ),
            config,
        )

    async def read_dtc_information_report_emissions_related_obd_dtc_by_status_mask(
        self,
        dtc_status_mask: int,
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportEmissionsRelatedOBDDTCByStatusMaskResponse,
    ]:
        """Read the number of emission related DTCs with the specified state from the UDS server.
        This is an implementation of the UDS request for the
        reportNumberOfEmissionsRelatedOBDDTCByStatusMask
        sub-function of the service ReadDTCInformation (0x19).

        :param dtc_status_mask: Used to select a portion of the DTCs based on their state.
        :param suppress_response: If set to True, the server is advised to not send back a positive
         response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReportEmissionsRelatedOBDDTCByStatusMaskRequest(
                dtc_status_mask, suppress_response
            ),
            config,
        )

    async def input_output_control_by_identifier(
        self,
        data_identifier: int,
        control_option_record: bytes,
        control_enable_mask_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        """Controls input or output values on the server.
        This is an implementation of the UDS request for the service InputOutputControlByIdentifier
        (0x2F).
        This function exposes the parameters as in the corresponding specification,
        hence is suitable for all variants of this service.
        For the variants which use an inputOutputControlParameter as the first byte of the
        controlOptionRecord, using the corresponding wrappers is recommended.

        :param data_identifier: The data identifier of the value(s) to be controlled.
        :param control_option_record: The controlStates, which specify the intended values of the
                                      input / output parameters, optionally prefixed with an
                                      inputOutputControlParameter or only an
                                      inputOutputControlParameter.
        :param control_enable_mask_record: In cases where the dataIdentifier corresponds to multiple
                                           input / output parameters, this mask specifies which ones
                                           should be affected by this request.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        pdu = struct.pack(
            "!BH", UDSIsoServices.InputOutputControlByIdentifier, data_identifier
        )
        pdu += control_option_record + control_enable_mask_record
        return await self.request(
            service.InputOutputControlByIdentifierRequest(
                data_identifier, control_option_record, control_enable_mask_record
            ),
            config,
        )

    async def input_output_control_by_identifier_return_control_to_ecu(
        self,
        data_identifier: int,
        control_enable_mask_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        """Gives the control over input / output parameters back to the ECU.
        This is a convenience wrapper for the generic input_output_control_by_id() for the case
        where an inputOutputControlParameter is used and is set to returnControlToECU.
        In that case no further controlState parameters can be submitted.

        :param data_identifier: The data identifier of the value(s) for which control should be
                                returned to the ECU.
        :param control_enable_mask_record: In cases where the dataIdentifier corresponds to multiple
                                           input / output parameters, this mask specifies which ones
                                           should be affected by this request.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ReturnControlToECURequest(
                data_identifier, control_enable_mask_record
            ),
            config,
        )

    async def input_output_control_by_identifier_reset_to_default(
        self,
        data_identifier: int,
        control_enable_mask_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        """Sets the input / output parameters to the default value(s).
        This is a convenience wrapper of the generic request for the case where an
        inputOutputControlParameter is used and is set to resetToDefault.
        In that case no further controlState parameters can be submitted.

        :param data_identifier: The data identifier of the value(s) for which the values should be
                                reset.
        :param control_enable_mask_record: In cases where the dataIdentifier corresponds to multiple
                                           input / output parameters, this mask specifies which ones
                                           should be affected by this request.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ResetToDefaultRequest(data_identifier, control_enable_mask_record),
            config,
        )

    async def input_output_control_by_identifier_freeze_current_state(
        self,
        data_identifier: int,
        control_enable_mask_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        """Freezes the input / output parameters at their current state.
        This is a convenience wrapper of the generic request for the case where an
        inputOutputControlParameter is used and is set to  freezeCurrentState.
        In that case no further controlState parameters can be submitted.

        :param data_identifier: The data identifier of the value(s) to be frozen.
        :param control_enable_mask_record: In cases where the dataIdentifier corresponds to multiple
                                           input / output parameters, this mask specifies which ones
                                           should be affected by this request.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.FreezeCurrentStateRequest(
                data_identifier, control_enable_mask_record
            ),
            config,
        )

    async def input_output_control_by_identifier_short_term_adjustment(
        self,
        data_identifier: int,
        control_states: bytes,
        control_enable_mask_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        """Sets the input / output parameters as specified in the controlOptionRecord.
        This is a convenience wrapper of the generic request for the case
        where an inputOutputControlParameter is used and is set to freezeCurrentState.
        In that case controlState parameters are required.

        :param data_identifier: The data identifier of the value(s) to be adjusted.
        :param control_states: The controlStates, which specify the intended values of the input /
                               output parameters.
        :param control_enable_mask_record: In cases where the dataIdentifier corresponds to multiple
                                           input / output parameters, this mask specifies which ones
                                           should be affected by this request.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.ShortTermAdjustmentRequest(
                data_identifier, control_enable_mask_record
            ),
            config,
        )

    async def routine_control_start_routine(
        self,
        routine_identifier: int,
        routine_control_option_record: bytes = b"",
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.StartRoutineResponse]:
        """Starts a specific routine on the server.
        This is an implementation of the UDS request for the startRoutine sub-function of the
        service routineControl (0x31).

        :param routine_identifier: The identifier of the routine.
        :param routine_control_option_record: Optional data.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.StartRoutineRequest(
                routine_identifier, routine_control_option_record, suppress_response
            ),
            config,
        )

    async def routine_control_stop_routine(
        self,
        routine_identifier: int,
        routine_control_option_record: bytes = b"",
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.StopRoutineResponse]:
        """Stops a specific routine on the server.
        This is an implementation of the UDS request for the stopRoutine sub-function of the service
        routineControl (0x31).

        :param routine_identifier: The identifier of the routine.
        :param routine_control_option_record: Optional data.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.StopRoutineRequest(
                routine_identifier, routine_control_option_record, suppress_response
            ),
            config,
        )

    async def routine_control_request_routine_results(
        self,
        routine_identifier: int,
        routine_control_option_record: bytes = b"",
        suppress_response: bool = False,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestRoutineResultsResponse]:
        """Requests the results of a specific routine on the server.
        This is an implementation of the UDS request for the requestRoutineResults sub-function of
        the service routineControl (0x31).

        :param routine_identifier: The identifier of the routine.
        :param routine_control_option_record: Optional data.
        :param suppress_response: If set to True, the server is advised to not send back a positive
                                  response.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.RequestRoutineResultsRequest(
                routine_identifier, routine_control_option_record, suppress_response
            ),
            config,
        )

    async def request_download(
        self,
        memory_address: int,
        memory_size: int,
        compression_method: int = 0x0,
        encryption_method: int = 0x0,
        address_and_length_format_identifier: Optional[int] = None,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestDownloadResponse]:
        """Requests the download of data, i.e. the possibility to send data from the client to the
        server.
        This is an implementation of the UDS request for requestDownload (0x34).

        :param memory_address: The address at which data should be downloaded.
        :param memory_size: The number of bytes to be downloaded.
        :param compression_method: Encodes the utilized compressionFormat (0x0 for none)
        :param encryption_method: Encodes the utilized encryptionFormat (0x0 for none)
        :param address_and_length_format_identifier: The byte lengths of the memory address and
                                                     size. If omitted, this parameter is computed
                                                     based on the memory_address and memory_size
                                                     parameters.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.RequestDownloadRequest(
                memory_address,
                memory_size,
                compression_method,
                encryption_method,
                address_and_length_format_identifier,
            ),
            config,
        )

    async def request_upload(
        self,
        memory_address: int,
        memory_size: int,
        compression_method: int = 0x0,
        encryption_method: int = 0x0,
        address_and_length_format_identifier: Optional[int] = None,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestUploadResponse]:
        """Requests the upload of data, i.e. the possibility to receive data from the server.
        This is an implementation of the UDS request for requestUpload (0x35).

        :param memory_address: The address at which data should be uploaded.
        :param memory_size: The number of bytes to be uploaded.
        :param compression_method: Encodes the utilized compressionFormat (0x0 for none)
        :param encryption_method: Encodes the utilized encryptionFormat (0x0 for none)
        :param address_and_length_format_identifier: The byte lengths of the memory address and
                                                     size. If omitted, this parameter is computed
                                                     based on the memory_address and memory_size
                                                     parameters.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.RequestUploadRequest(
                memory_address,
                memory_size,
                compression_method,
                encryption_method,
                address_and_length_format_identifier,
            ),
            config,
        )

    async def transfer_data(
        self,
        block_sequence_counter: int,
        transfer_request_parameter_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.TransferDataResponse]:
        """Transfers data to the server or requests the next data from the server.
        This is an implementation of the UDS request for transferData (0x36).

        :param block_sequence_counter: The current block sequence counter.
                                       Initialized with one and incremented for each new data.
                                       After 0xff, the counter is resumed at 0
        :param transfer_request_parameter_record: Contains the data to be transferred if downloading
                                                  to the server.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.TransferDataRequest(
                block_sequence_counter, transfer_request_parameter_record
            ),
            config,
        )

    async def request_transfer_exit(
        self,
        transfer_request_parameter_record: bytes = b"",
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestTransferExitResponse]:
        """Ends the transfer of data.
        This is an implementation of the UDS request for requestTransferExit (0x77).

        :param transfer_request_parameter_record: Optional data.
        :param config: Passed on to request_pdu().
        :return: The response of the server.
        """
        return await self.request(
            service.RequestTransferExitRequest(transfer_request_parameter_record),
            config,
        )

    @overload
    async def request(
        self, request: service.RawRequest, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.PositiveResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.DiagnosticSessionControlRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.DiagnosticSessionControlResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ECUResetRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ECUResetResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.RequestSeedRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.SecurityAccessResponse]:
        ...

    @overload
    async def request(
        self, request: service.SendKeyRequest, config: Optional[UDSRequestConfig] = None
    ) -> Union[service.NegativeResponse, service.SecurityAccessResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.CommunicationControlRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.CommunicationControlResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.TesterPresentRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.TesterPresentResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ControlDTCSettingRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ControlDTCSettingResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ReadDataByIdentifierRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReadDataByIdentifierResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ReadMemoryByAddressRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReadMemoryByAddressResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.WriteDataByIdentifierRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.WriteDataByIdentifierResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.WriteMemoryByAddressRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.WriteMemoryByAddressResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ClearDiagnosticInformationRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ClearDiagnosticInformationResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportNumberOfDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReportNumberOfDTCByStatusMaskResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.ReportDTCByStatusMaskResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportMirrorMemoryDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.ReportMirrorMemoryDTCByStatusMaskResponse
    ]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportNumberOfMirrorMemoryDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportNumberOfMirrorMemoryDTCByStatusMaskResponse,
    ]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportNumberOfEmissionsRelatedOBDDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportNumberOfEmissionsRelatedOBDDTCByStatusMaskResponse,
    ]:
        ...

    @overload
    async def request(
        self,
        request: service.ReportEmissionsRelatedOBDDTCByStatusMaskRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse,
        service.ReportEmissionsRelatedOBDDTCByStatusMaskResponse,
    ]:
        ...

    @overload
    async def request(
        self,
        request: service.InputOutputControlByIdentifierRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[
        service.NegativeResponse, service.InputOutputControlByIdentifierResponse
    ]:
        ...

    @overload
    async def request(
        self,
        request: service.StartRoutineRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.StartRoutineResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.StopRoutineRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.StopRoutineResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.RequestRoutineResultsRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestRoutineResultsResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.RequestDownloadRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestDownloadResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.RequestUploadRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestUploadResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.TransferDataRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.TransferDataResponse]:
        ...

    @overload
    async def request(
        self,
        request: service.RequestTransferExitRequest,
        config: Optional[UDSRequestConfig] = None,
    ) -> Union[service.NegativeResponse, service.RequestTransferExitResponse]:
        ...

    async def request(
        self, request: service.UDSRequest, config: Optional[UDSRequestConfig] = None
    ) -> service.UDSResponse:
        """Sends a raw UDS request and returns the response.
        Network errors are handled via exponential backoff.
        Pending errors, triggered by the ECU are resolved as well.

        :param request: request to send
        :param config: The request config parameters
        :return: The response.
        """
        return await self._request(request, config)

    async def _request(
        self, request: service.UDSRequest, config: Optional[UDSRequestConfig] = None
    ) -> service.UDSResponse:
        async with self.transport.mutex:
            return await self.request_unsafe(request, config)



import asyncio
import sys
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse
from gallia.uds.ecu import ECU
from gallia.transports.base import TargetURI
from gallia.transports.can import ISOTPTransport, RawCANTransport
from gallia.udscan.core import DiscoveryScanner
from gallia.udscan.utils import auto_int, write_ecu_url_list


class FindCanIDsScanner(DiscoveryScanner):
    """This scanner discovers all UDS endpoints on a ECU using ISO-TP normal addressing.
    This is the default protocol used by OBD.
    When using normal addressing, the ISO-TP header does not include an address and there is no generic tester address.
    Addressing is only done via CAN IDs. Every endpoint has a source and destination CAN ID.
    Typically, there is also a broadcast destination ID to address all endpoints."""

    def add_parser(self) -> None:
        self.parser.set_defaults(tester_present=False)

        self.parser.add_argument(
            "--start",
            metavar="INT",
            type=auto_int,
            default=0x00,
            help="set start address",
        )
        self.parser.add_argument(
            "--end",
            metavar="INT",
            type=auto_int,
            default=0x7FF,
            help="set end address",
        )
        self.parser.add_argument(
            "--info-did",
            metavar="DID",
            type=auto_int,
            default=0xF197,
            help="DID to query ECU description",
        )
        self.parser.add_argument(
            "--sniff-time",
            default=60,
            type=int,
            metavar="SECONDS",
            help="Time in seconds to sniff on bus for current traffic",
        )

    async def setup(self, args: Namespace) -> None:
        if not args.target.scheme == RawCANTransport.SCHEME:
            self.logger.log_error(
                f"Unsupported transport schema {args.target.scheme}; must be can-raw!"
            )
            sys.exit(1)
        await super().setup(args)

    async def main(self, args: Namespace) -> None:
        assert isinstance(self.transport, RawCANTransport)
        found = []
        can_url = args.target.url._replace(scheme=ISOTPTransport.SCHEME)

        sniff_time: int = args.sniff_time
        self.logger.log_summary(
            f"Listening to idle bus communication for {sniff_time}s..."
        )
        addr_idle = await self.transport.get_idle_traffic(sniff_time)
        self.logger.log_summary(f"Found {len(addr_idle)} CAN Addresses on idle Bus")
        self.transport.set_filter(addr_idle, inv_filter=True)

        for ID in range(args.start, args.end + 1):
            self.logger.log_info(f"Testing CAN ID {ID:03x}")
            is_broadcast = False

            await self.transport.sendto(
                bytes([0x02, 0x10, 0x01, 0xAA, 0xAA, 0xAA, 0xAA, 0xAA]),
                timeout=0.1,
                dst=ID,
            )
            try:
                addr, _ = await self.transport.recvfrom(timeout=0.1)
                if addr == ID:
                    self.logger.log_info(
                        f"wtf!? The same CAN ID {ID:03x} answered. Skipping..."
                    )
                    continue
            except asyncio.TimeoutError:
                continue

            while True:
                # The recv buffer needs to be flushed to avoid
                # wrong results...
                try:
                    new_addr, _ = await self.transport.recvfrom(timeout=0.1)
                    if new_addr != addr:
                        is_broadcast = True
                        self.logger.log_summary(
                            f"seems that broadcast was triggered on CAN ID {ID:03x}, "
                            f"got answer from {new_addr:03x}"
                        )
                    else:
                        self.logger.log_info(
                            f"seems like a large ISO-TP packet was received on CAN ID {ID:03x}"
                        )
                except asyncio.TimeoutError:
                    if is_broadcast:
                        self.logger.log_summary(
                            f"seems that broadcast was triggered on CAN ID {ID:03x}, "
                            f"got answer from {addr:03x}"
                        )
                    else:
                        self.logger.log_summary(
                            f"found endpoint on CAN ID [src:dst]: {ID:03x}:{addr:03x}"
                        )
                        found.append(
                            (
                                can_url,
                                {
                                    "src_addr": hex(ID),
                                    "dst_addr": hex(addr),
                                    "tx_padding": "0xaa",
                                    "rx_padding": "0xaa",
                                    "is_fd": str(self.transport.args["is_fd"]).lower(),
                                },
                            )
                        )
                    break

        self.logger.log_summary(f"finished; found {len(found)} UDS endpoints")
        ecus_file = self.artifacts_dir.joinpath("ECUs.txt")
        self.logger.log_summary(f"Writing urls to file: {ecus_file}")
        connection_strings = await write_ecu_url_list(ecus_file, found, self.db_handler)

        self.logger.log_info("reading info DID from all discovered endpoints")
        did = args.info_did
        for url in connection_strings:
            self.logger.log_summary("----------------------------")
            self.logger.log_summary(f"Probing ECU: {url}")
            target = TargetURI(url)
            transport = ISOTPTransport(target)
            await transport.connect(None)
            ecu = ECU(transport, timeout=2)
            self.logger.log_summary(f"reading device description at 0x{did:04x}")
            try:
                resp = await ecu.read_data_by_identifier(did)
                if isinstance(resp, NegativeResponse):
                    self.logger.log_summary(f"could not read did: {resp}")
                else:
                    self.logger.log_summary(f"response was: {resp.data_record!r}")
            except Exception as e:
                self.logger.log_summary(
                    f"reading description failed: {e.__class__.__name__} {e}"
                )
                await asyncio.sleep(0.1)



import asyncio
import sys
from argparse import Namespace

from gallia.transports.can import RawCANTransport, ISOTPTransport
from gallia.udscan.core import DiscoveryScanner
from gallia.udscan.utils import write_ecu_url_list


class FindISOTPAddrScanner(DiscoveryScanner):
    """This scanner discovers all UDS endpoints on a ECU using ISO-TP extended addressing.
    For this addressing schema, a tester-id is used as source CAN-ID for all endpoints.
    Different endpoints are addressed via ISO-TP extended address (1 byte).
    This scanner iterates all ISO-TP extended addresses to find all endpoints."""

    def add_parser(self) -> None:
        self.parser.set_defaults(tester_present=False)

        self.parser.add_argument(
            "--sniff-time",
            default=60,
            type=int,
            metavar="SECONDS",
            help="Time in seconds to sniff on bus for current traffic",
        )

    async def setup(self, args: Namespace) -> None:
        if not args.target.scheme == RawCANTransport.SCHEME:
            self.logger.log_error(
                f"Unsupported transport schema {args.target.scheme}; must be can-raw!"
            )
            sys.exit(1)
        if "src_addr" not in args.target.qs:
            self.logger.log_error("CAN source ID (test-id) must be set!")
            sys.exit(1)
        await super().setup(args)

    async def main(self, args: Namespace) -> None:
        assert isinstance(self.transport, RawCANTransport)
        found = []
        url = args.target.url._replace(scheme=ISOTPTransport.SCHEME)

        sniff_time: int = args.sniff_time
        self.logger.log_summary(
            f"Listening to idle bus communication for {sniff_time}s..."
        )
        addr_idle = await self.transport.get_idle_traffic(sniff_time)
        self.logger.log_summary(f"Found {len(addr_idle)} CAN Addresses on idle Bus")
        self.transport.set_filter(addr_idle, inv_filter=True)
        # flush receive queue
        await self.transport.get_idle_traffic(2)

        self.logger.log_summary("Starting with search")
        tester_id = int(args.target.qs["src_addr"][0], 0)
        for ID in range(0x100):
            self.logger.log_info(f"Testing ISO-TP address: {ID:02x}")
            is_broadcast = False

            await self.transport.sendto(
                bytes([ID, 0x02, 0x10, 0x01]), tester_id, timeout=0.1
            )
            try:
                addr_old, data = await self.transport.recvfrom(timeout=0.2)
                if addr_old == ID:
                    self.logger.log_warning(
                        f"{ID:02x}: wtf!? The same ID answered. Skipping..."
                    )
                    continue
            except asyncio.TimeoutError:
                continue

            new_addr = None
            while True:
                # The recv buffer needs to be flushed to avoid
                # wrong results...
                try:
                    new_addr, _ = await self.transport.recvfrom(timeout=0.2)
                    if new_addr != addr_old:
                        if not is_broadcast:
                            self.logger.log_summary(
                                f"found broadcast endpoint on ID {ID:02x}; "
                                f"response from CAN ID [src:dst] {tester_id:03x}:{addr_old:03x} "
                                f"ISO-TP [tx:rx] {ID:02x}:{data[0]:02x}"
                            )
                        self.logger.log_summary(
                            f"found broadcast endpoint on ID {ID:02x}; "
                            f"response from CAN ID [src:dst] {tester_id:03x}:{new_addr:03x} "
                            f"ISO-TP [tx:rx] {ID:02x}:{data[0]:02x}"
                        )
                        is_broadcast = True
                    else:
                        self.logger.log_summary(
                            f"{tester_id:03x}:{new_addr:03x}: seems like "
                            f"a large ISO-TP packet was received on ISO-TP address: {ID:02x}"
                        )
                except asyncio.TimeoutError:
                    if new_addr is None:
                        if is_broadcast is False and data is not None:
                            msg = (
                                f"found endpoint: CAN [src:dst]: {tester_id:03x}:{addr_old:03x} "
                                f"ISO-TP [tx:rx]: {data[0]:02x}:{ID:02x}, entire payload: {data.hex()}"
                            )
                            self.logger.log_summary(msg)
                            found.append(
                                (
                                    url,
                                    {
                                        "src_addr": hex(tester_id),
                                        "dst_addr": hex(addr_old),
                                        "rx_ext_address": hex(data[0]),
                                        "ext_address": hex(ID),
                                        "is_fd": str(
                                            self.transport.args["is_fd"]
                                        ).lower(),
                                    },
                                )
                            )
                    break

        self.logger.log_summary(f"Found {len(found)} ISO-TP endpoints")
        ecus_file = self.artifacts_dir.joinpath("ECUs.txt")
        self.logger.log_summary(f"Writing urls to file: {ecus_file}")
        await write_ecu_url_list(ecus_file, found, self.db_handler)



# pylint: disable=too-many-lines,eval-used

from __future__ import annotations

import curses
import curses.ascii
import json
import gzip
import mmap
import shutil
import tempfile
import warnings
from argparse import ArgumentParser
from array import array
from binascii import unhexlify
from copy import deepcopy
from dataclasses import dataclass
from datetime import datetime
from enum import IntEnum
from math import ceil
from pathlib import Path
from typing import Optional, Any, BinaryIO, Union

import zstandard as zstd
from penlog import MessagePrio, RecordType

from gallia.uds.core.service import UDSResponse, UDSRequest, NegativeResponse


class InterpretationColor(IntEnum):
    DEFAULT = 42
    UDS_REQUEST = 43
    UDS_POSITIVE_RESPONSE = 44
    UDS_NEGATIVE_RESPONSE = 45


@dataclass
class PenlogEntry(RecordType):
    interpretation: Optional[str] = None
    interpretation_color: InterpretationColor = InterpretationColor.DEFAULT


@dataclass
class FormattedText:
    text: str
    format: int


@dataclass
class DisplayEntry:
    """
    Represents one single line of a single formatted penlog entry as displayed in the console.

    :param texts: The (differently) formatted text fragments of the displayed line.
    :param penlog_entry_number: The line number of the corresponding penlog entry.
    :param entry_line_number: The index in the list of formatted lines as created for the corresponding penlog entry.
    :param last_line: True if and only if this entry refers to the last of the formatted lines as created for the
                      corresponding penlog entry.
    """

    texts: list[FormattedText]
    penlog_entry_number: int
    entry_line_number: int
    last_line = False


@dataclass
class PriorityZone:
    """
    Represents a continuous range (including start and end) of entries for which a certain priority filter applies.

    :param start: The first penlog entry of the zone.
    :param end: The last penlog entry of the zone.
    :param priority: The filtering priority to be applied on all entries in in the zone.
    """

    start: int
    end: Optional[int]
    priority: MessagePrio


class EntryCache:
    """
    A simple two level cache that stores the latest accessed penlog entries.

    This is used to speed up the processing of penlog entries by reducing the number of necessary fetches from the
    input file. It may be ineffective if the whole file is kept in memory.
    """

    def __init__(
        self,
        file: Union[BinaryIO, mmap.mmap],
        entry_positions: array,
        cache_size: int = 20_000,
    ):
        self.file = file
        self.old_entries: dict[int, PenlogEntry] = {}
        self.entries: dict[int, PenlogEntry] = {}
        self.entry_positions = entry_positions
        self.cache_size = cache_size

    def __len__(self) -> int:
        return len(self.entry_positions)

    def __getitem__(self, entry_number: int) -> PenlogEntry:
        try:
            return self.entries[entry_number]
        except KeyError:
            self._load_entry(entry_number)
            return self.entries[entry_number]

    def _load_entry(self, entry_number: int) -> None:
        self.file.seek(self.entry_positions[entry_number])
        line = self.file.readline()

        if len(self.entries) + len(self.old_entries) >= self.cache_size:
            self.old_entries = self.entries
            self.entries = {}

        try:
            self.entries[entry_number] = self.old_entries.pop(entry_number)
        except KeyError:
            try:
                self.entries[entry_number] = PenlogEntry(**json.loads(line))
            except (json.decoder.JSONDecodeError, TypeError):
                self.entries[entry_number] = PenlogEntry(
                    data=line.decode("utf-8"),
                    host="",
                    component="JSON",
                    timestamp="",
                    type="ERROR",
                    priority=MessagePrio.ERROR,
                )


@dataclass
class Configuration:
    priority_zones: list[PriorityZone]
    filter: list[str]
    interpret: bool


class CursedHR:
    """
    A curses based interactive reader for log files complying to penlog.

    It supports the following features:
        - interactive display of penlog files (similar to less)
        - coloring and formatting of entries based on the priority level
        - optimized sectional filtering based on priority
        - unoptimized sectional filtering based on any penlog entry attribute
        - in-line interpretation of UDS messages using the Gallia UDS implementation

    See https://github.com/Fraunhofer-AISEC/penlog for the penlog specification.
    """

    def __init__(
        self,
        in_file: Path,
        priority: MessagePrio = MessagePrio.DEBUG,
        filters: Optional[list[str]] = None,
    ):
        self.in_file = in_file
        self.level_pointers: list[array] = list(array("L") for _ in range(9))
        self.entry_positions = array("L")
        self.configuration_history = [
            Configuration(
                priority_zones=[PriorityZone(0, None, priority)],
                filter=filters if filters is not None else [],
                interpret=False,
            )
        ]
        self.configuration_index = 0

        try:
            self.window = self.init_curses()
            self.color_ids = self.define_colors()
            self.file = self.uncompressed_file()

            try:
                with mmap.mmap(
                    self.file.fileno(), 0, access=mmap.ACCESS_READ
                ) as mm_file:
                    self.parse_structure(mm_file)
                    self.entries = EntryCache(mm_file, self.entry_positions)
                    self.handle_io()
            finally:
                self.file.close()
        finally:
            self.terminate_curses()

    def init_curses(self) -> Any:
        window = curses.initscr()
        curses.start_color()
        curses.use_default_colors()
        curses.noecho()
        curses.cbreak()
        window.keypad(True)

        return window

    def terminate_curses(self) -> None:
        curses.nocbreak()
        self.window.keypad(False)
        curses.echo()
        curses.endwin()

    def define_colors(self) -> dict:
        prio_colors = {
            MessagePrio.EMERGENCY: (100, curses.COLOR_RED),
            MessagePrio.ALERT: (101, curses.COLOR_RED),
            MessagePrio.CRITICAL: (102, curses.COLOR_RED),
            MessagePrio.ERROR: (103, curses.COLOR_RED),
            MessagePrio.WARNING: (104, curses.COLOR_YELLOW),
            MessagePrio.NOTICE: (105, -1),
            MessagePrio.INFO: (106, -1),
            MessagePrio.DEBUG: (107, 8),
            MessagePrio.TRACE: (108, curses.COLOR_BLUE),
        }

        for (identifier, value) in prio_colors.values():
            curses.init_pair(identifier, value, -1)

        curses.init_pair(InterpretationColor.DEFAULT, 8, -1)
        curses.init_pair(InterpretationColor.UDS_REQUEST, curses.COLOR_CYAN, -1)
        curses.init_pair(
            InterpretationColor.UDS_POSITIVE_RESPONSE, curses.COLOR_GREEN, -1
        )
        curses.init_pair(InterpretationColor.UDS_NEGATIVE_RESPONSE, 166, -1)

        return dict((prio, color[0]) for prio, color in prio_colors.items())

    def uncompressed_file(self) -> BinaryIO:
        """
        Returns an uncompressed version of the input file as referenced by self.in_file.

        In case the input file is already uncompressed, the original file is returned.
        In any other case, the input file is decompressed to a temporary file (potentially on disk).
        """
        self.window.clear()
        self.window.addstr(f"Loading contents from {self.in_file}")
        self.window.refresh()

        try:
            if self.in_file.suffix in [".zst", ".gz"]:
                self.window.erase()
                self.window.addstr(
                    f"Loading contents from {self.in_file}: Decompressing file ..."
                )
                self.window.refresh()

                file = tempfile.TemporaryFile()  # pylint: disable=consider-using-with

                if self.in_file.suffix == ".zst":
                    with self.in_file.open("rb") as in_file:
                        decomp = zstd.ZstdDecompressor()
                        decomp.copy_stream(in_file, file)
                elif self.in_file.suffix == ".gz":
                    with gzip.open(self.in_file, "rb") as in_file:
                        shutil.copyfileobj(in_file, file)
            else:
                file = self.in_file.open("rb")

            return file  # type: ignore
        except Exception as e:
            raise ValueError("Unsupported file format") from e

    def parse_structure(self, file: Union[BinaryIO, mmap.mmap]) -> None:
        """
        Parses an (already uncompressed) penlog file to a skeleton containing structural information but no data.

        This process enables handling relatively large penlog entries efficiently by extracting information on the
        priority from each entry, which can later be used to quickly jump to adjacent entries of at least same priority.
        This is accomplished by the following two data structures:

            - self.entry_positions: Stores the absolute offsets of each penlog entry inside the input file.
                                    This offset can later be used to efficiently retrieve the data for a penlog entry
                                    from the input file without the need to store it in memory.
            - self.level_pointers: Stores the entries which have at least a certain priority (i.e. less or equal).
                                   More precisely, the indices of the corresponding entries in self.entry_positions.
                                   There exists on such list for each priority level.
                                   This allows to quickly find entries which are to be displayed inside a priority zone
                                   even if there are many entries with filtered out priority in between.
        """
        prio_prefix = '"priority":'.encode("ascii")
        prio_prefix_len = len(prio_prefix)

        file.seek(0, 2)
        file_length = file.tell()
        file.seek(0)

        self.window.erase()
        self.window.addstr(
            f"Loading contents from {self.in_file}: Parsing structure ({0}%)"
        )
        self.window.refresh()

        prev_progress = 0
        num_entries = 0

        while True:
            n = file.tell()
            line = file.readline()

            if not line:
                break

            prio = MessagePrio.INFO.value

            if line[0] != 123:
                prio = MessagePrio.ERROR
            elif (pos := line.find(prio_prefix)) > 0:
                pos += prio_prefix_len

                while not 48 <= line[pos] <= 57:
                    pos += 1

                prio = line[pos] - 48

            for i in reversed(range(prio, MessagePrio.TRACE + 1)):
                self.level_pointers[i].append(num_entries)

            self.entry_positions.append(n)
            num_entries += 1

            progress = n * 100 // file_length

            if progress >= prev_progress + 10:
                prev_progress = progress
                self.window.erase()
                self.window.addstr(
                    f"Loading contents from {self.in_file}: Parsing structure ({progress}%)"
                )
                self.window.refresh()

    @property
    def configuration(self) -> Configuration:
        return self.configuration_history[self.configuration_index]

    def new_configuration(self) -> None:
        self.configuration_index += 1
        self.configuration_history = self.configuration_history[
            : self.configuration_index
        ]
        self.configuration_history.append(deepcopy(self.configuration_history[-1]))

    def entry_zone(self, entry_id: int) -> PriorityZone:
        """
        Returns the zone into which the entry with the given id falls.

        :params entry_id: The index of the entry.
        :return: The zone into which the entry with the given id falls.
        """
        for zone in self.configuration.priority_zones:
            if entry_id >= zone.start and (zone.end is None or entry_id <= zone.end):
                return zone

        assert False

    def update_zones(self, new_zone: PriorityZone) -> None:
        """
        Inserts the given zone into the current set of zones.

        Adds a new list of zones into the zone history, which contains the previous zones and the new zone.
        In the process of integrating the new zone, itself as well as old zones may be altered or completely removed.
        It is guaranteed, that the zones are always ordered and do not overlap.

        :param new_zone: The zone to be added.
        """
        assert new_zone.end is not None

        self.new_configuration()

        i = 0

        while i < len(self.configuration.priority_zones) - 1:
            cur_zone = self.configuration.priority_zones[i]
            assert cur_zone.end is not None

            # If no intersection
            if new_zone.end < cur_zone.start or new_zone.start > cur_zone.end:
                i += 1
                continue

            # if current is contained in new
            if new_zone.start <= cur_zone.start and new_zone.end >= cur_zone.end:
                self.configuration.priority_zones.remove(cur_zone)
                continue

            # if new is contained in current
            if new_zone.start > cur_zone.start and new_zone.end < cur_zone.end:
                self.configuration.priority_zones.insert(
                    i + 1,
                    PriorityZone(new_zone.end + 1, cur_zone.end, cur_zone.priority),
                )
                cur_zone.end = new_zone.start - 1
                break

            # if new is intersecting from the left side
            if new_zone.end < cur_zone.end:
                cur_zone.start = new_zone.end + 1

            # if new is intersecting from the right side
            if new_zone.start > cur_zone.start:
                cur_zone.end = new_zone.start - 1

            i += 1

        # if new is contained in or intersecting from the left side into the last zone
        last_zone = self.configuration.priority_zones[-1]

        if new_zone.end >= last_zone.start:
            if new_zone.start >= last_zone.start:
                self.configuration.priority_zones.insert(
                    -1,
                    PriorityZone(
                        last_zone.start, new_zone.start - 1, last_zone.priority
                    ),
                )
            last_zone.start = new_zone.end + 1

        i = 0

        while new_zone.start > self.configuration.priority_zones[i].start:
            i += 1

        self.configuration.priority_zones.insert(i, new_zone)

    def format_text(self, text: str, entry: PenlogEntry) -> FormattedText:
        """
        Returns a formatted text according to the properties of the corresponding penlog entry.

        :param text: The text to be formatted.
        :param entry: The penlog entry to which the text corresponds.
        :return: The formatted text.
        """
        if entry.component == "JSON":
            return FormattedText(
                text, curses.color_pair(self.color_ids[MessagePrio.ERROR])
            )

        text_format = curses.color_pair(self.color_ids[entry.priority])

        if entry.priority < MessagePrio.INFO:
            text_format = text_format | curses.A_BOLD

        return FormattedText(text, text_format)

    def default_text(self, text: str) -> FormattedText:
        """
        Returns a formatted text with the content of the given text and the default formatting.

        :param text: The text.
        :return: The formatted text.
        """
        return FormattedText(text, curses.color_pair(0))

    def formatted_entry(self, entry_id: int) -> list[DisplayEntry]:
        """
        Returns a list of formatted entries, each corresponding to a single line of the entry referred to by the
        given entry id, as displayed in the console.

        :param entry_id: The index of the entry.
        :return: The formatted display entries for the given entry.
        """
        entry = self.entries[entry_id]
        _, max_width = self.window.getmaxyx()

        try:
            formatted_timestamp = datetime.fromisoformat(
                entry.timestamp[:26] + entry.timestamp[-6:]
            ).strftime("%b %e %H:%M:%S.%f")[:-3]
        except ValueError:
            formatted_timestamp = "0" * 19

        prefix = f"{formatted_timestamp} {{{entry.component:8.8}}} [{entry.type:8.8}]: "
        residual_width = max_width - len(prefix) - 1

        user_defined_lines = entry.data.splitlines()

        if len(user_defined_lines) == 0:
            user_defined_lines = [""]

        terminal_width_defined_lines = []

        for line in user_defined_lines:
            if len(line) == 0:
                terminal_width_defined_lines.append("")

            for i in range(ceil(len(line) / residual_width)):
                terminal_width_defined_lines.append(
                    line[i * residual_width : (i + 1) * residual_width]
                )

        result = [
            DisplayEntry(
                [
                    self.default_text(prefix),
                    self.format_text(terminal_width_defined_lines[0], entry),
                ],
                entry_id,
                0,
            )
        ]

        for i, line in enumerate(terminal_width_defined_lines[1:]):
            result.append(
                DisplayEntry(
                    [
                        self.default_text(" " * len(prefix)),
                        self.format_text(line, entry),
                    ],
                    entry_id,
                    i + 1,
                )
            )

        if self.configuration.interpret:
            if entry.interpretation is None:
                self.interpret_entry(entry)

            if entry.interpretation is not None:
                interpretation_text = f"  # {entry.interpretation}"

                while len(interpretation_text) > 0:
                    len_texts = 0

                    for text in result[-1].texts:
                        len_texts += len(text.text)

                    residual_width = max_width - len_texts - 1

                    result[-1].texts.append(
                        FormattedText(
                            interpretation_text[:residual_width],
                            curses.color_pair(entry.interpretation_color),
                        )
                    )

                    interpretation_text = interpretation_text[residual_width:]

                    if len(interpretation_text) > 0:
                        result.append(
                            DisplayEntry(
                                [self.default_text(" " * len(prefix))],
                                entry_id,
                                result[-1].entry_line_number,
                            )
                        )

        result[-1].last_line = True

        return result

    def interpret_entry(self, entry: PenlogEntry) -> None:
        """
        Adds an interpretation to the entry, if the entry can be identified as referring to a UDS request or response.

        :param entry: The entry to be interpreted.
        """
        try:
            if not entry.data.startswith("00"):
                data = unhexlify(entry.data)

                if data[0] & 0b01000000:
                    response = UDSResponse.parse_dynamic(data)
                    entry.interpretation = repr(response)

                    if isinstance(response, NegativeResponse):
                        entry.interpretation_color = (
                            InterpretationColor.UDS_NEGATIVE_RESPONSE
                        )
                    else:
                        entry.interpretation_color = (
                            InterpretationColor.UDS_POSITIVE_RESPONSE
                        )
                else:
                    entry.interpretation = repr(UDSRequest.parse_dynamic(data))
                    entry.interpretation_color = InterpretationColor.UDS_REQUEST
        except Exception as e:
            self.debug_log(repr(e))

    def check_filter(self, entry_id: int) -> bool:
        """
        Check if the entry which is referred to by the given entry id, is still to be displayed after applying filters.

        :param entry_id: The index of the entry.
        :return: The formatted display entries for the given entry.
        """
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", SyntaxWarning)

            for command in self.configuration.filter:
                if not eval(command, self.entries[entry_id].__dict__):
                    return False

        return True

    def calculate_display_entries(
        self, start_entry: int, entry_line: int
    ) -> list[DisplayEntry]:
        """
        Returns a list of display entries starting from the given start entry, which qualify to be displayed by having
        a sufficient priority as well as passing any other filter.

        The display entries may start from a different entry than the one pointed to by the given parameter,
        in case that entry itself does not qualify itself.
        In that case previous entries are taken into account if the entry_line is set to -1
        and if there is at least oneprevious entry which qualifies.
        Otherwise only entries following the given one are taken into account.

        The number of entries which is returned is limited by the number of lines in the console.

        The list is guaranteed to be not empty.
        If there would be no entries to be displayed there will be an info message,
        which informs the user, that no entries are available with the current filtering settings.

        :param start_entry: The starting point for the search for the actual first entry (which qualifies) in the list.
        :param entry_line: The index in the list of formatted lines as created for the corresponding penlog entry.
                           Most importantly, a value of 0 indicates to start from the first line,
                           a value of -1 to start from the last line. Other negative values are not supported.
        :return: The list of display entries.
        """
        max_lines, _ = self.window.getmaxyx()
        max_lines -= 1
        display_entries: list[DisplayEntry] = []

        prio = self.entries[start_entry].priority
        zone = self.entry_zone(start_entry)

        fallback_entry = [
            DisplayEntry(
                [
                    self.default_text(
                        "No entries found! You can select a new priority on this line"
                    )
                ],
                0,
                0,
            )
        ]

        if prio > zone.priority or not self.check_filter(start_entry):
            if entry_line == -1:
                start_entry_ = self.previous_sufficient_entry(start_entry)

                if start_entry_ is None:
                    start_entry_ = self.next_sufficient_entry(0)

                    if start_entry_ is None:
                        return fallback_entry

                    entry_line = 0

                start_entry = start_entry_
            elif entry_line == 0:
                start_entry_ = self.next_sufficient_entry(start_entry)

                if start_entry_ is None:
                    return fallback_entry

                start_entry = start_entry_

            zone = self.entry_zone(start_entry)

        pointer = self.priority_pointer(start_entry, zone.priority)
        assert pointer is not None, f"{start_entry} {zone.priority}"

        i = 0

        while True:
            entry_id = self.level_pointers[zone.priority][pointer]
            new_lines = self.formatted_entry(entry_id)

            if i == 0:
                if entry_line == -1:
                    display_entries.append(new_lines[-1])
                else:
                    display_entries += new_lines[entry_line:]
            else:
                display_entries += new_lines

            if len(display_entries) > max_lines:
                break
            i += 1

            zone, pointer = self.next_sufficient_pointer(zone, pointer, entry_id)

            if pointer is None:
                break

        return display_entries[:max_lines]

    def next_sufficient_entry(self, entry_id: int) -> Optional[int]:
        """
        Returns the index of the next entry after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters.

        If no such entry exists None is returned.

        :param entry_id: The index of the entry in self.entries.
        :return: The index of the next sufficient entry after the given one or None, if no such entry exists.
        """
        prio = self.entries[entry_id].priority
        pointer = self.priority_pointer(entry_id, prio)
        assert pointer is not None
        zone, pointer = self.next_sufficient_pointer(None, pointer, entry_id)
        return (
            self.level_pointers[zone.priority][pointer] if pointer is not None else None
        )

    def previous_sufficient_entry(self, entry_id: int) -> Optional[int]:
        """
        Returns the index of the previous entry after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters.

        If no such entry exists None is returned.

        :param entry_id: The index of the entry in self.entries.
        :return: The index of the previous sufficient entry before the given one or None, if no such entry exists.
        """
        prio = self.entries[entry_id].priority
        pointer = self.priority_pointer(entry_id, prio)
        assert pointer is not None
        zone, pointer = self.previous_sufficient_pointer(None, pointer, entry_id)
        return (
            self.level_pointers[zone.priority][pointer] if pointer is not None else None
        )

    def next_sufficient_pointer(
        self, prio_zone: Optional[PriorityZone], pointer: int, entry_id: int
    ) -> tuple[PriorityZone, Optional[int]]:
        """
        Returns the index of the next priority pointer after the given one,
        which has a sufficiently high priority to be displayed and also passes any filters,
        along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The next priority pointer's index along the corresponding priority zone.
        """
        zone, pointer = self.next_sufficient_pointer_without_filters(
            prio_zone, pointer, entry_id
        )

        while pointer is not None and not self.check_filter(
            self.level_pointers[zone.priority][pointer]
        ):
            entry_id = self.level_pointers[zone.priority][pointer]
            zone, pointer = self.next_sufficient_pointer_without_filters(
                prio_zone, pointer, entry_id
            )

        return zone, pointer

    def previous_sufficient_pointer(
        self, prio_zone: Optional[PriorityZone], pointer: int, entry_id: int
    ) -> tuple[PriorityZone, Optional[int]]:
        """
        Returns the index of the previous priority pointer before the given one,
        which has a sufficiently high priority to be displayed and also passes any filters,
        along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The previous priority pointer's index along the corresponding priority zone.
        """
        zone, pointer = self.previous_sufficient_pointer_without_filters(
            prio_zone, pointer, entry_id
        )

        while pointer is not None and not self.check_filter(
            self.level_pointers[zone.priority][pointer]
        ):
            entry_id = self.level_pointers[zone.priority][pointer]
            zone, pointer = self.previous_sufficient_pointer_without_filters(
                prio_zone, pointer, entry_id
            )

        return zone, pointer

    def next_sufficient_pointer_without_filters(
        self, prio_zone: Optional[PriorityZone], pointer: int, entry_id: int
    ) -> tuple[PriorityZone, Optional[int]]:
        """
        Returns the index of the next priority pointer after the given one,
        which has a sufficiently high priority to be displayed, along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The next priority pointer's index along the corresponding priority zone.
        """
        for zone in self.configuration.priority_zones:
            if zone.end is not None and entry_id >= zone.end:
                continue

            prio_entries = self.level_pointers[zone.priority]

            if len(prio_entries) == 0:
                continue

            next_pointer: Optional[int]

            if zone == prio_zone and pointer + 1 < len(prio_entries):
                next_pointer = pointer + 1
            else:
                next_pointer = self.priority_pointer(
                    max(entry_id, zone.start - 1), zone.priority, 1
                )

            if next_pointer is not None and (
                zone.end is None or prio_entries[next_pointer] <= zone.end
            ):
                return zone, next_pointer

        return self.configuration.priority_zones[-1], None

    def previous_sufficient_pointer_without_filters(
        self, prio_zone: Optional[PriorityZone], pointer: int, entry_id: int
    ) -> tuple[PriorityZone, Optional[int]]:
        """
        Returns the index of the previous priority pointer before the given one,
        which has a sufficiently high priority to be displayed, along with the priority zone to which it belongs.

        If no such pointer exists None is returned for the pointer.

        :param prio_zone: The zone into which the entry, to which the given pointer points, lies.
        :param pointer: The given pointer's index.
        :param entry_id: The index of the entry, to which the given pointer points, in self.entries.
        :return: The previous priority pointer's index along the corresponding priority zone.
        """
        for zone in reversed(self.configuration.priority_zones):
            if entry_id <= zone.start:
                continue

            prio_entries = self.level_pointers[zone.priority]

            if len(prio_entries) == 0:
                continue

            prev_pointer: Optional[int]

            if zone == prio_zone and pointer > 0:
                prev_pointer = pointer - 1
            else:
                prev_pointer = self.priority_pointer(
                    min(entry_id, zone.end + 1) if zone.end is not None else entry_id,
                    zone.priority,
                    -1,
                )

            if prev_pointer is not None and prio_entries[prev_pointer] >= zone.start:
                return zone, prev_pointer

        return self.configuration.priority_zones[0], None

    def priority_pointer(
        self, entry_id: int, prio: MessagePrio, mode: int = 0
    ) -> Optional[int]:
        """
        Returns the index of the level pointer of the entry with the given id,
        or depending on the mode the previous or next entry, in the level pointers with the given priority.

        Depending on the mode parameter different pointer indices can be retrieved:
         - case 0: The pointer which points to the exact same entry which is given as the entry id
         - case 1: - 1: The pointer with the next bigger entry_id in the level pointers compared to the given entry id
         - case -1: The pointer with the next smaller entry_id in the level pointers compared to the given entry id

         In none of these cases it can be guaranteed that such a pointer exists in the level pointers for the given
         priority. If no pointer exists, None is returned.

        :param entry_id: The index of the entry in self.entries.
        :param prio: The priority level for which the index in the corresponding level pointers should be calculated.
        :param mode: Determines which pointer index should be returned (must be any of 0, 1 or -1).
        :return: The index of the level pointer or None if no such pointer could be identified.
        """
        prio_entries = self.level_pointers[prio]

        upper = len(prio_entries) - 1
        lower = 0

        while True:
            pointer = (upper + lower) // 2

            if upper == lower + 1 and mode == -1:
                pointer = upper

            prio_entry = prio_entries[pointer]

            if upper == lower:
                break

            # Find the exact match
            if mode == 0:
                if prio_entry > entry_id:
                    upper = max(lower, pointer - 1)
                elif prio_entry < entry_id:
                    lower = min(pointer + 1, upper)
                else:
                    break

            # Find the next bigger entry
            if mode == 1:
                if prio_entry > entry_id:
                    upper = max(lower, pointer)
                else:
                    lower = min(pointer + 1, upper)

            # Find the next smaller entry
            if mode == -1:
                if prio_entry >= entry_id:
                    upper = max(lower, pointer - 1)
                else:
                    lower = min(pointer, upper)

        if mode == 0 and prio_entry != entry_id:
            return None

        if mode == 1 and prio_entry <= entry_id:
            return None

        if mode == -1 and prio_entry >= entry_id:
            return None

        return pointer

    def handle_io(self) -> None:
        start_entry: Optional[int] = None
        max_lines, max_columns = self.window.getmaxyx()
        max_lines -= 1
        filter_history = [self.configuration.filter]

        display_entries = self.calculate_display_entries(0, 0)
        self.display(
            display_entries, self.status(display_entries, (len(display_entries) - 1, 0))
        )
        prefix_length = 42
        cursor = (0, prefix_length)
        self.window.move(*cursor)

        def update_selected_zones(prio: MessagePrio) -> None:
            nonlocal start_entry
            nonlocal display_entries
            nonlocal cursor
            nonlocal max_lines

            if start_entry is None:
                return

            stop_display_entry = display_entries[
                min(cursor[0], len(display_entries) - 1)
            ]
            stop_entry = stop_display_entry.penlog_entry_number

            if start_entry == stop_entry:
                start_entry = self.previous_sufficient_entry(start_entry)
                stop_entry_ = self.next_sufficient_entry(stop_entry)

                if start_entry is None:
                    start_entry = 0

                if stop_entry_ is None:
                    stop_entry = len(self.entries) - 1
                else:
                    stop_entry = stop_entry_

            self.update_zones(
                PriorityZone(
                    min(start_entry, stop_entry), max(start_entry, stop_entry), prio
                )
            )
            start_entry = None

        while (key := self.window.getkey()) != "q":
            entry_start = display_entries[0].penlog_entry_number
            line_start = display_entries[0].entry_line_number
            cursor = curses.getsyx()
            max_lines, max_columns = self.window.getmaxyx()
            max_lines -= 1

            def page_up() -> None:
                nonlocal entry_start
                nonlocal line_start
                nonlocal display_entries

                for _ in range(max_lines - 1):
                    old_entry_start = entry_start
                    old_line_start = line_start

                    if display_entries[0].entry_line_number == 0 and entry_start > 0:
                        entry_start = max(0, display_entries[0].penlog_entry_number - 1)
                        line_start = -1
                    else:
                        line_start = display_entries[0].entry_line_number - 1

                    if old_entry_start == entry_start and old_line_start == line_start:
                        break

                    display_entries = self.calculate_display_entries(
                        entry_start, line_start
                    )

            def line_up() -> None:
                nonlocal entry_start
                nonlocal line_start

                if display_entries[0].entry_line_number == 0 and entry_start > 0:
                    entry_start = max(0, entry_start - 1)
                    line_start = -1
                else:
                    line_start -= 1

            if key == "KEY_UP":
                if cursor[0] > 0:
                    cursor = (cursor[0] - 1, cursor[1])
                else:
                    line_up()
            elif key == "KEY_DOWN":
                if cursor[0] < max_lines - 1:
                    cursor = cursor[0] + 1, cursor[1]
                else:
                    if display_entries[0].last_line:
                        entry_start = min(len(self.entries) - 1, entry_start + 1)
                        line_start = 0
                    else:
                        line_start += 1
            elif key == "KEY_PPAGE":
                if cursor[0] > 0:
                    cursor = (0, cursor[1])
                else:
                    page_up()
            elif key == "KEY_NPAGE":
                if cursor[0] < len(display_entries) - 1:
                    cursor = (len(display_entries) - 1, cursor[1])
                else:
                    entry_start = display_entries[-1].penlog_entry_number
                    line_start = display_entries[-1].entry_line_number
            elif key == "g":
                entry_start = 0
                line_start = 0
                cursor = (0, cursor[1])
            elif key == "G":
                entry_start = len(self.entries) - 1
                line_start = -1

                display_entries = self.calculate_display_entries(
                    entry_start, line_start
                )
                page_up()
                cursor = (len(display_entries) - 1, cursor[1])
            elif key == "KEY_LEFT":
                if cursor[1] > 0:
                    self.window.move(cursor[0], cursor[1] - 1)
                    continue
            elif key == "KEY_RIGHT":
                if cursor[1] < max_columns - 1:
                    self.window.move(cursor[0], cursor[1] + 1)
                    continue
            elif key in ["v", "V"]:
                if cursor[0] < len(display_entries):
                    start_entry = display_entries[cursor[0]].penlog_entry_number
            elif key == chr(curses.ascii.ESC):
                start_entry = None
            elif key in ["p", "P"]:
                function_key = key

                while (key := self.window.getkey()) != "q":
                    if key == chr(curses.ascii.ESC):
                        start_entry = None
                        break
                    elif key == "m":
                        prio = MessagePrio.EMERGENCY
                    elif key == "a":
                        prio = MessagePrio.ALERT
                    elif key == "c":
                        prio = MessagePrio.CRITICAL
                    elif key == "e":
                        prio = MessagePrio.ERROR
                    elif key == "w":
                        prio = MessagePrio.WARNING
                    elif key == "n":
                        prio = MessagePrio.NOTICE
                    elif key == "i":
                        prio = MessagePrio.INFO
                    elif key == "d":
                        prio = MessagePrio.DEBUG
                    elif key == "t":
                        prio = MessagePrio.TRACE
                    else:
                        continue

                    if function_key == "p":
                        update_selected_zones(prio)
                    else:
                        self.new_configuration()
                        self.configuration.priority_zones = [
                            PriorityZone(0, None, prio)
                        ]

                    break
            elif key == "u":
                self.configuration_index = max(0, self.configuration_index - 1)
            elif key == "r":
                self.configuration_index = min(
                    len(self.configuration_history) - 1, self.configuration_index + 1
                )
            elif key == "i":
                self.new_configuration()
                self.configuration.interpret = not self.configuration.interpret
            elif key == "f":
                # fh is short for filter_history to reduce long unreadable lines
                fh_tmp = list(
                    "; ".join(filter_commands) for filter_commands in filter_history
                )
                fh_tmp.append("")
                fh_index = len(fh_tmp) - 1

                filter_cursor = len(fh_tmp[fh_index])
                input_format = curses.color_pair(0)

                self.display(
                    display_entries, [FormattedText(fh_tmp[fh_index], input_format)]
                )
                self.window.move(max_lines, filter_cursor)

                while (key := self.window.getkey()) != chr(curses.ascii.ESC):
                    if key == "\n":
                        try:
                            filter_tmp = parse_filter(fh_tmp[fh_index])
                            self.debug_log(filter_tmp)
                            filter_history.append(filter_tmp)
                            self.new_configuration()
                            self.configuration.filter = filter_tmp
                            break
                        except Exception:
                            pass

                    elif key == "KEY_BACKSPACE":
                        if filter_cursor > 0:
                            fh_tmp[fh_index] = (
                                fh_tmp[fh_index][: filter_cursor - 1]
                                + fh_tmp[fh_index][filter_cursor:]
                            )
                            filter_cursor -= 1
                    elif key == "KEY_DC":
                        if filter_cursor < len(fh_tmp[fh_index]):
                            fh_tmp[fh_index] = (
                                fh_tmp[fh_index][:filter_cursor]
                                + fh_tmp[fh_index][filter_cursor + 1 :]
                            )
                    elif key == "KEY_LEFT":
                        if filter_cursor > 0:
                            filter_cursor -= 1
                    elif key == "KEY_RIGHT":
                        if filter_cursor < len(fh_tmp[fh_index]):
                            filter_cursor += 1
                    elif key == "KEY_UP":
                        fh_index = max(0, fh_index - 1)
                    elif key == "KEY_DOWN":
                        fh_index = min(len(fh_tmp) - 1, fh_index + 1)
                    else:
                        fh_tmp[fh_index] = (
                            fh_tmp[fh_index][:filter_cursor]
                            + key
                            + fh_tmp[fh_index][filter_cursor:]
                        )
                        filter_cursor += 1

                    try:
                        parse_filter(fh_tmp[fh_index])
                        input_format = curses.color_pair(0)
                    except Exception:
                        input_format = curses.color_pair(
                            self.color_ids[MessagePrio.WARNING]
                        )

                    self.display(
                        display_entries, [FormattedText(fh_tmp[fh_index], input_format)]
                    )
                    self.window.move(
                        max_lines, min(filter_cursor, len(fh_tmp[fh_index]))
                    )
            else:
                pass

            display_entries = self.calculate_display_entries(entry_start, line_start)

            previous_entry_start = 0
            previous_line_start = 0

            while len(display_entries) < max_lines and (
                previous_entry_start != entry_start or previous_line_start != line_start
            ):
                previous_entry_start = entry_start
                previous_line_start = line_start
                entry_start = display_entries[0].penlog_entry_number
                line_start = display_entries[0].entry_line_number
                line_up()
                display_entries = self.calculate_display_entries(
                    entry_start, line_start
                )

            if start_entry is not None:
                stop_entry = display_entries[
                    min(cursor[0], len(display_entries) - 1)
                ].penlog_entry_number

                for entry in display_entries:
                    if (
                        min(start_entry, stop_entry)
                        <= entry.penlog_entry_number
                        <= max(start_entry, stop_entry)
                    ):
                        for text in entry.texts:
                            text.format = text.format | curses.A_REVERSE

            self.display(display_entries, self.status(display_entries, cursor))
            self.window.move(min(cursor[0], len(display_entries) - 1), cursor[1])

    def status(
        self, display_entries: list[DisplayEntry], cursor: tuple[int, int]
    ) -> list[FormattedText]:
        selected_entry = display_entries[min(cursor[0], len(display_entries) - 1)]
        entry_number = selected_entry.penlog_entry_number + 1
        n = len(self.entries)
        zone = self.entry_zone(selected_entry.penlog_entry_number)
        zone_info = self.default_text(
            f"Zone {self.configuration.priority_zones.index(zone): >2} ({zone.priority.name})"
        )
        progress = self.default_text(
            f"{entry_number: >{len(str(n))}}/{n} ({(entry_number / n) * 100:6.2f}%)"
        )
        _, max_columns = self.window.getmaxyx()
        spacing = self.default_text(
            " " * (max_columns - 1 - len(zone_info.text) - len(progress.text))
        )
        return [zone_info, spacing, progress]

    def display(
        self, display_entries: list[DisplayEntry], status_line: list[FormattedText]
    ) -> None:
        self.window.erase()

        for i, display_entry in enumerate(display_entries):
            for text in display_entry.texts:
                self.window.addstr(text.text, text.format)

            if i < len(display_entries) - 1:
                self.window.addstr("\n")

        self.window.move(self.window.getmaxyx()[0] - 1, 0)

        for text in status_line:
            self.window.addstr(text.text, text.format)

    def debug_log(self, msg: Any) -> None:
        debug_path = Path("/tmp/cursed_log")

        try:
            with debug_path.open("a", encoding="utf-8") as f:
                f.write(str(msg) + "\n")
        except Exception:
            pass


def parse_prio(text: str) -> MessagePrio:
    try:
        return MessagePrio(int(text, 0))
    except ValueError:
        try:
            return MessagePrio[text.upper()]
        except ValueError as e:
            raise ValueError("Not a valid priority") from e


def parse_filter(text: str) -> list[str]:
    test_entry = PenlogEntry(
        "component",
        "data",
        "host",
        MessagePrio.INFO,
        "1970-01-01T00:00:00.000000+00:00",
        "type",
    )

    commands = list(
        command.strip() for command in text.split(";") if len(command.strip()) > 0
    )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", SyntaxWarning)

        for command in commands:
            eval(command, test_entry.__dict__)

    return commands


def main() -> None:
    parser = ArgumentParser()
    parser.add_argument("file", type=Path)
    parser.add_argument("--priority", "-p", type=parse_prio, default=MessagePrio.DEBUG)
    parser.add_argument("--filter", "-f", type=parse_filter, default=None)
    args = parser.parse_args()
    CursedHR(args.file, args.priority, args.filter)


if __name__ == "__main__":
    main()



import asyncio
from argparse import Namespace, Action, ArgumentError, ArgumentParser
from subprocess import run
from sys import stdout
from pathlib import Path
from urllib.parse import urlencode, ParseResult
from typing import Any, Callable, Optional, Union, Sequence

import aiofiles

from gallia.db.db_handler import DBHandler
from gallia.uds.ecu import ECU
from gallia.penlog import Logger
from gallia.uds.core.exception import UnexpectedNegativeResponse
from gallia.uds.core.client import UDSRequestConfig
from gallia.uds.helpers import suggests_identifier_not_supported
from gallia.uds.core import service


def auto_int(arg: str) -> int:
    return int(arg, 0)


async def poll_service(
    ecu: ECU, sid: int, check: Optional[int] = None
) -> Union[service.NegativeResponse, service.PositiveResponse]:
    if check:
        cur_session = await ecu.read_session()
        if cur_session != check:
            ecu.logger.log_warning("ecu lost session")
            ecu.logger.log_warning(f"expected: {check:x}; read: {cur_session:x}")
            await ecu.set_session(check)

    resp = await ecu.send_raw(bytes([sid]))

    if check:
        cur_session = await ecu.read_session()
        if cur_session != check:
            await ecu.set_session(check)
            ecu.logger.log_warning(f"expected: {check:x}; read: {cur_session:x}")
    return resp


async def find_sessions(ecu: ECU, search: list, max_retry: int = 4) -> list[int]:
    sessions = []
    for sid in search:
        try:
            resp = await ecu.set_session(
                sid, config=UDSRequestConfig(max_retry=max_retry)
            )
            if isinstance(resp, service.NegativeResponse):
                continue
        except Exception:
            continue
        sessions.append(sid)
        await ecu.leave_session(sid)
    return sessions


async def check_and_set_session(
    ecu: ECU, expected_session: int, retries: int = 3
) -> bool:  # pylint: disable=R0911
    """check_and_set_session() reads the current session and (re)tries to set
    the session to the expected session if they do not match.

    Returns True if the current session matches the expected session,
    or if read_session is not supported by the ECU or in the current session."""

    ecu.logger.log_debug(
        f"Checking current session, expecting 0x{expected_session:02x}"
    )

    try:
        current_session = await ecu.read_session(
            config=UDSRequestConfig(max_retry=retries)
        )
    except UnexpectedNegativeResponse as e:
        if suggests_identifier_not_supported(e.RESPONSE_CODE):
            ecu.logger.log_info(
                f"Read current session not supported: {e.RESPONSE_CODE.name}, skipping check_session"
            )
            return True
        raise e
    except asyncio.TimeoutError:
        ecu.logger.log_warning(
            "Reading current session timed out, skipping check_session"
        )
        return True

    ecu.logger.log_debug(f"Current session is 0x{current_session:02x}")
    if current_session == expected_session:
        return True

    for i in range(retries):
        ecu.logger.log_warning(
            f"Not in session 0x{expected_session:02x}, ECU replied with 0x{current_session:02x}"
        )

        ecu.logger.log_info(
            f"Switching to session 0x{expected_session:02x}; attempt {i + 1} of {retries}"
        )
        resp = await ecu.set_session(expected_session)

        if isinstance(resp, service.NegativeResponse):
            ecu.logger.log_warning(
                f"Switching to session 0x{expected_session:02x} failed: {resp}"
            )

        try:
            current_session = await ecu.read_session(
                config=UDSRequestConfig(max_retry=retries)
            )
            ecu.logger.log_debug(f"Current session is 0x{current_session:02x}")
            if current_session == expected_session:
                return True
        except UnexpectedNegativeResponse as e:
            if suggests_identifier_not_supported(e.RESPONSE_CODE):
                ecu.logger.log_info(
                    f"Read current session not supported: {e.RESPONSE_CODE.name}, skipping check_session"
                )
                return True
            raise e
        except asyncio.TimeoutError:
            ecu.logger.log_warning(
                "Reading current session timed out, skipping check_session"
            )
            return True

    ecu.logger.log_warning(
        f"Failed to switch to session 0x{expected_session:02x} after {retries} attempts"
    )
    return False


def get_command_output(command: list[str]) -> str:
    """get_command_output runs the command given as a list and returns a string
    of the corresponding output from STDOUT. If any error occurs, the error
    string is returned instead.

    :param command: A list of strings without whitespace which specify the command
    :return: A string of the command's output from STDOUT, or an error string"""
    try:
        process = run(command, capture_output=True, check=True)
    except Exception as e:
        return f"Error: {e}"

    return process.stdout.decode().strip()


def unravel(listing: str) -> list[int]:
    listing_delimiter = ","
    range_delimiter = "-"
    result = set()

    for range_element in listing.split(listing_delimiter):
        if range_delimiter in range_element:
            first_tmp, last_tmp = range_element.split(range_delimiter)
            first = auto_int(first_tmp)
            last = auto_int(last_tmp)

            for element in range(first, last + 1):
                result.add(element)
        else:
            element = auto_int(range_element)
            result.add(element)

    return sorted(result)


class ParseSkips(Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: str = None,
    ) -> None:
        skip_sids: dict[int, Optional[list[int]]] = {}

        try:
            if values is not None:
                for session_skips in values:
                    # Whole sessions can be skipped by only giving the session number without ids
                    if ":" not in session_skips:
                        session_ids = unravel(session_skips)

                        for session_id in session_ids:
                            skip_sids[session_id] = None
                    else:
                        session_ids_tmp, identifier_ids_tmp = session_skips.split(":")
                        session_ids = unravel(session_ids_tmp)
                        identifier_ids = unravel(identifier_ids_tmp)

                        for session_id in session_ids:
                            if session_id not in skip_sids:
                                skip_sids[session_id] = []

                            session_skips = skip_sids[session_id]

                            if session_skips is not None:
                                session_skips += identifier_ids

            setattr(namespace, self.dest, skip_sids)
        except Exception as e:
            raise ArgumentError(self, "The argument is malformed!") from e


class ANSIEscapes:
    if stdout.isatty():
        BOLD = "\033[1m"
        ITALIC = "\033[3m"
        UNDERSCORE = "\033[4m"
        BLINK = "\033[5m"
        CROSSED = "\033[9m"

        BLACK = "\033[90m"
        RED = "\033[91m"
        GREEN = "\033[92m"
        YELLOW = "\033[93m"
        BLUE = "\033[94m"
        MAGENTA = "\033[95m"
        CYAN = "\033[96m"
        WHITE = "\033[97m"

        RESET = "\033[0m"
    else:
        BOLD = ITALIC = UNDERSCORE = BLINK = CROSSED = ""
        BLACK = RED = GREEN = YELLOW = BLUE = MAGENTA = CYAN = WHITE = ""
        RESET = ""


def range_diff(base: list[range], coverage: list[range]) -> list[range]:
    """This function computes ranges of the base, that are not contained
    in the coverage.

    The main idea of this algorithm is to go through each range of the base and to iteratively shrink it by
    removing the ranges in the coverage. What's left is then the uncovered area.
    There are in principle five cases which are considered for a pair of ranges:

    - One or both are empty
    - They are disjoint
      -> In both cases nothing is removed from the base range
    - The base range is a subrange of the coverage range
      -> The base range becomes empty
    - The coverage range reaches into base range
      -> The base range gets smaller
    - The coverage range is a subrange of the base range
      -> The base range is split into two smaller ranges

    :param base: The base ranges.
    :param coverage: The coverage ranges.
    :return: The (sub-)ranges of the base ranges that are not covered by the coverage ranges.
    """
    result: list[range] = []
    sorted_base = sorted(base, key=lambda r: r[0])
    i = 0

    # We are looping with a while statement, because we may insert new elements into the list
    while i < len(sorted_base):
        base_r = sorted_base[i]

        for cover_r in sorted(coverage, key=lambda r: r[0]):
            # None is empty and they are not disjoint
            if not (
                len(base_r) == 0
                or len(cover_r) == 0
                or cover_r[-1] < base_r[0]
                or cover_r[0] > base_r[-1]
            ):
                # The base range is a subrange of the coverage range
                if cover_r[0] <= base_r[0] and cover_r[-1] >= base_r[-1]:
                    base_r = range(0)
                # The coverage range reaches into base range from the left
                elif cover_r[0] <= base_r[0]:
                    base_r = range(cover_r[-1] + 1, base_r[-1] + 1)
                # The coverage range reaches into base range from the right
                elif cover_r[-1] >= base_r[-1]:
                    base_r = range(base_r[0], cover_r[0])
                # The coverage range is a subrange of the base range
                # In this case the base range is divided by the coverage range into two ranges.
                # The first of them will take the place of the current base range and the second will be handled
                # separately in another iteration.
                else:
                    sorted_base.insert(i + 1, range(cover_r[-1] + 1, base_r[-1] + 1))
                    base_r = range(base_r[0], cover_r[0])

        # If an uncovered range is left after all elements of the coverage have been removed from the base range
        if len(base_r) > 0:
            result.append(base_r)

        i += 1

    return result


async def write_ecu_url_list(
    ecus_file: Path,
    ecus: list[tuple[ParseResult, dict[str, Any]]],
    db_handler: Optional[DBHandler] = None,
) -> list[str]:
    """Write a list of ECU connection strings (urls) into file

    :param ecus_file: output file
    :param ecus: list of ECUs with ECU specific url and params as dict
    :params db_handler: if given, urls are also written to the database as discovery results
    :return: None
    """
    urls: list[str] = list()
    async with aiofiles.open(ecus_file, "w") as file:
        for url, ecu in ecus:
            url_result = f"{url.scheme}://{url.netloc}?{urlencode(ecu)}"
            urls.append(url_result)
            await file.write(f"{url_result}\n")

            if db_handler is not None:
                await db_handler.insert_discovery_result(url_result)
    return urls


async def catch_and_log_exception(
    logger: Logger, func: Callable, *args: Any, **kwargs: Any
) -> None:
    """Runs an async function. If an exception is raised,
    it will be logged via logger.

    :param logger: an instance of gallia.penlog.Logger
    :param func: a async function object which will be awaited
    :return: None
    """
    try:
        return await func(*args, **kwargs)
    except Exception as e:
        logger.log_error(f"func {func.__name__} failed: {repr(e)}")



import asyncio
from argparse import Namespace

from gallia.uds.core.constants import UDSErrorCodes
from gallia.uds.core.service import (
    DiagnosticSessionControlResponse,
    DiagnosticSessionControlRequest,
    NegativeResponse,
)
from gallia.uds.helpers import raise_for_mismatch
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int, write_ecu_url_list


class FindEndpoints(UDSScanner):
    """This is a generic UDS endpoint discovery scanner. Currently only supports DoIP."""

    def __init__(self) -> None:
        super().__init__()

        self.implicit_logging = False
        self.log_scan_run = False

    def add_parser(self) -> None:
        self.parser.set_defaults(tester_present=False)

        self.parser.add_argument(
            "--reversed",
            action="store_true",
            help="scan in reversed order",
        )
        self.parser.add_argument(
            "--start",
            metavar="INT",
            type=auto_int,
            default=0x00,
            help="set start address",
        )
        self.parser.add_argument(
            "--end", metavar="INT", type=auto_int, default=0xFF, help="set end address"
        )

    async def setup(self, args: Namespace) -> None:
        await super().setup(args)

        if self.db_handler is not None:
            try:
                await self.db_handler.insert_discovery_run(args.target.url.scheme)
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the discovery run to the database: {repr(e)}"
                )

    async def main(self, args: Namespace) -> None:
        if self.db_handler is not None:
            try:
                await self.db_handler.insert_discovery_run(args.target.url.scheme)
            except Exception as e:
                self.logger.log_warning(
                    f"Could not write the discovery run to the database: {repr(e)}"
                )

        assert self.transport is not None
        src_addr = self.transport.args["src_addr"]  # type: ignore
        found = []
        src_gen = (
            range(args.end + 1, args.start)
            if args.reversed
            else range(args.start, args.end + 1)
        )
        req = DiagnosticSessionControlRequest(0x01)
        url = args.target.url

        for dst_addr in src_gen:
            try:
                await self.transport.sendto(req.pdu, dst_addr)
                while True:
                    addr, data = await self.transport.recvfrom(timeout=args.timeout)
                    resp = DiagnosticSessionControlResponse.parse_static(data)
                    raise_for_mismatch(req, resp)
                    if (
                        isinstance(resp, NegativeResponse)
                        and resp.response_code
                        == UDSErrorCodes.requestCorrectlyReceivedResponsePending
                    ):
                        continue
                    break
                self.logger.log_info(f"{resp}")
                self.logger.log_info(f"found dst: {addr:2x}")
                found.append(
                    (url, {"dst_addr": hex(dst_addr), "src_addr": hex(src_addr)})
                )
            except BrokenPipeError:
                await self.transport.reconnect()
            except asyncio.TimeoutError:
                pass

        self.logger.log_summary(f"Finished scan; found {len(found)} endpoints")
        ecus_file = self.artifacts_dir.joinpath("ECUs.txt")
        self.logger.log_summary(f"Writing urls to file: {ecus_file}")
        await write_ecu_url_list(ecus_file, found, self.db_handler)



from enum import Enum
from typing import Any, Optional

import penlog


class MessageType(str, Enum):
    READ = "read"
    WRITE = "write"
    PREAMBLE = "preamble"
    MESSAGE = "message"
    SUMMARY = "summary"


class Logger(penlog.Logger):
    stack_depth = 4

    def log_preamble(self, data: Any) -> None:
        self.log_msg(
            data,
            MessageType.PREAMBLE,
            penlog.MessagePrio.INFO,
            _depth=self.stack_depth,
        )

    def log_read(self, data: Any, tags: Optional[list[str]] = None) -> None:
        self.log_msg(
            data,
            MessageType.READ,
            penlog.MessagePrio.DEBUG,
            tags,
            _depth=self.stack_depth,
        )

    def log_write(self, data: Any, tags: Optional[list[str]] = None) -> None:
        self.log_msg(
            data,
            MessageType.WRITE,
            penlog.MessagePrio.DEBUG,
            tags,
            _depth=self.stack_depth,
        )

    def log_summary(self, data: Any, tags: Optional[list[str]] = None) -> None:
        self.log_msg(
            data,
            MessageType.SUMMARY,
            penlog.MessagePrio.NOTICE,
            tags,
            _depth=self.stack_depth,
        )


class DiscardLogger(Logger):
    def _log(self, msg: penlog.RecordType, depth: int) -> None:
        pass



import asyncio
from argparse import Namespace

from gallia.uds.core.service import NegativeResponse, UDSResponse
from gallia.udscan.core import UDSScanner
from gallia.udscan.utils import auto_int


class EcuReset(UDSScanner):
    """ECU_Reset"""

    def add_parser(self) -> None:
        self.parser.set_defaults(properties=False)

        self.parser.add_argument(
            "--session", type=auto_int, default=0x01, help="set session perform test in"
        )
        self.parser.add_argument(
            "-f", "--subfunc", type=auto_int, default=0x01, help="subfunc"
        )

    async def main(self, args: Namespace) -> None:
        resp: UDSResponse = await self.ecu.set_session(args.session)
        if isinstance(resp, NegativeResponse):
            self.logger.log_error(f"could not change to session: 0x{args.session:02x}")
            return

        try:
            self.logger.log_info(f"try sub-func: 0x{args.subfunc:02x}")
            resp = await self.ecu.ecu_reset(args.subfunc)
            if isinstance(resp, NegativeResponse):
                msg = f"ECU Reset 0x{args.subfunc:02x} failed in session: 0x{args.session:02x}: {resp}"
                self.logger.log_error(msg)
            else:
                self.logger.log_summary(f"ECU Reset 0x{args.subfunc:02x} succeeded")
        except asyncio.TimeoutError:
            self.logger.log_error("Timeout")
            await asyncio.sleep(10)
        except ConnectionError:
            msg = f"Lost connection to ECU, session: 0x{args.session:02x} subFunc: 0x{args.subfunc:02x}"
            self.logger.log_error(msg)
            return



from argparse import Namespace

from gallia.udscan.core import Scanner
from gallia.udscan.utils import catch_and_log_exception
from gallia.services.xcp import XCPService


class TestXCP(Scanner):
    """Test XCP Slave"""

    def add_parser(self) -> None:
        pass

    async def main(self, args: Namespace) -> None:
        assert self.transport
        service = XCPService(self.transport)

        await catch_and_log_exception(self.logger, service.connect)
        await catch_and_log_exception(self.logger, service.get_status)
        await catch_and_log_exception(self.logger, service.get_comm_mode_info)
        await catch_and_log_exception(self.logger, service.disconnect)



from __future__ import annotations

import asyncio
from abc import ABC
from typing import Optional, Any, Type

from gallia.uds.core.service import UDSRequest, UDSResponse, NegativeResponse
from gallia.uds.core.constants import UDSErrorCodes


# ****************
# * Base classes *
# ****************


class UDSException(Exception):
    def __init__(self, request: UDSRequest, message: Optional[str] = None):
        self.request = request
        self.message = message

        super().__init__(message)

    def _message_core(self) -> str:
        return f"triggered by {self.request}"

    def __str__(self) -> str:
        message = self._message_core()

        if self.message is not None:
            message = f"{message}; {self.message}"

        return message

    def __repr__(self) -> str:
        return f"{type(self).__name__}({repr(str(self))})"


class MissingResponse(UDSException, asyncio.TimeoutError):
    def _message_core(self) -> str:
        return f"triggered by {self.request}"


class ResponseException(UDSException):
    def __init__(
        self, request: UDSRequest, response: UDSResponse, message: Optional[str] = None
    ):
        self.response = response

        super().__init__(request, message)

    def _message_core(self) -> str:
        return f"{self.response} to {self.request}"


class IllegalResponse(ResponseException):
    pass


class RequestResponseMismatch(IllegalResponse):
    def _message_core(self) -> str:
        return f"{repr(self.response)} to {self.request}"


class MalformedResponse(IllegalResponse):
    pass


class UnexpectedResponse(ResponseException):
    pass


class UnexpectedNegativeResponse(UnexpectedResponse, ABC):
    RESPONSE_CODE: UDSErrorCodes
    _CONCRETE_EXCEPTIONS: dict[
        Optional[UDSErrorCodes], Type[UnexpectedNegativeResponse]
    ] = {}

    def __init_subclass__(cls, /, response_code: UDSErrorCodes, **kwargs: Any) -> None:
        super().__init_subclass__(**kwargs)

        cls.RESPONSE_CODE = response_code
        UnexpectedNegativeResponse._CONCRETE_EXCEPTIONS[response_code] = cls

    def __init__(
        self,
        request: UDSRequest,
        response: NegativeResponse,
        message: Optional[str] = None,
    ):
        self.response: NegativeResponse = response

        super().__init__(request, response, message)

    @staticmethod
    def parse_dynamic(
        request: UDSRequest, response: NegativeResponse, message: Optional[str] = None
    ) -> UnexpectedNegativeResponse:
        return UnexpectedNegativeResponse._CONCRETE_EXCEPTIONS[response.response_code](
            request, response, message
        )


# ******************************************
# * Concrete Unexpected Negative Responses *
# ******************************************

# Auto-generated using the following code snippet:
# for ec in UDSErrorCodes:
#     print(f'class {ec.name[0].upper()}{ec.name[1:]}('
#           f'\n        UnexpectedNegativeResponse, response_code={str(ec)}):\n    pass\n\n')


class GeneralReject(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.generalReject
):
    pass


class ServiceNotSupported(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.serviceNotSupported
):
    pass


class SubFunctionNotSupported(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.subFunctionNotSupported
):
    pass


class IncorrectMessageLengthOrInvalidFormat(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.incorrectMessageLengthOrInvalidFormat,
):
    pass


class ResponseTooLong(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.responseTooLong
):
    pass


class BusyRepeatRequest(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.busyRepeatRequest
):
    pass


class ConditionsNotCorrect(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.conditionsNotCorrect
):
    pass


class RequestSequenceError(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.requestSequenceError
):
    pass


class NoResponseFromSubnetComponent(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.noResponseFromSubnetComponent,
):
    pass


class RequestOutOfRange(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.requestOutOfRange
):
    pass


class SecurityAccessDenied(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.securityAccessDenied
):
    pass


class AuthenticationRequired(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.authenticationRequired
):
    pass


class InvalidKey(UnexpectedNegativeResponse, response_code=UDSErrorCodes.invalidKey):
    pass


class ExceededNumberOfAttempts(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.exceededNumberOfAttempts
):
    pass


class RequiredTimeDelayNotExpired(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.requiredTimeDelayNotExpired
):
    pass


class SecureDataTransmissionRequired(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.secureDataTransmissionRequired,
):
    pass


class SecureDataTransmissionNotAllowed(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.secureDataTransmissionNotAllowed,
):
    pass


class SecureDataVerificationFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.secureDataVerificationFailed
):
    pass


class CertificateVerificationFailedInvalidTimePeriod(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidTimePeriod,
):
    pass


class CertificateVerificationFailedInvalidSignature(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidSignature,
):
    pass


class CertificateVerificationFailedInvalidChainOfTrust(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidChainOfTrust,
):
    pass


class CertificateVerificationFailedInvalidType(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidType,
):
    pass


class CertificateVerificationFailedInvalidFormat(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidFormat,
):
    pass


class CertificateVerificationFailedInvalidContent(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidContent,
):
    pass


class CertificateVerificationFailedInvalidScope(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidScope,
):
    pass


class CertificateVerificationFailedInvalidCertificateRevoked(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.certificateVerificationFailedInvalidCertificateRevoked,
):
    pass


class OwnershipVerificationFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.ownershipVerificationFailed
):
    pass


class ChallengeCalculationFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.challengeCalculationFailed
):
    pass


class SettingAccessRightsFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.settingAccessRightsFailed
):
    pass


class SessionKeyCreationOrDerivationFailed(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.sessionKeyCreationOrDerivationFailed,
):
    pass


class ConfigurationDataUsageFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.configurationDataUsageFailed
):
    pass


class DeAuthenticationFailed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.deAuthenticationFailed
):
    pass


class UploadDownloadNotAccepted(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.uploadDownloadNotAccepted
):
    pass


class TransferDataSuspended(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.transferDataSuspended
):
    pass


class GeneralProgrammingFailure(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.generalProgrammingFailure
):
    pass


class WrongBlockSequenceCounter(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.wrongBlockSequenceCounter
):
    pass


class RequestCorrectlyReceivedResponsePending(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.requestCorrectlyReceivedResponsePending,
):
    pass


class SubFunctionNotSupportedInActiveSession(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.subFunctionNotSupportedInActiveSession,
):
    pass


class ServiceNotSupportedInActiveSession(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.serviceNotSupportedInActiveSession,
):
    pass


class RpmTooHigh(UnexpectedNegativeResponse, response_code=UDSErrorCodes.rpmTooHigh):
    pass


class RpmTooLow(UnexpectedNegativeResponse, response_code=UDSErrorCodes.rpmTooLow):
    pass


class EngineIsRunning(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.engineIsRunning
):
    pass


class EngineIsNotRunning(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.engineIsNotRunning
):
    pass


class EngineRunTimeTooLow(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.engineRunTimeTooLow
):
    pass


class TemperatureTooHigh(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.temperatureTooHigh
):
    pass


class TemperatureTooLow(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.temperatureTooLow
):
    pass


class VehicleSpeedTooHigh(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.vehicleSpeedTooHigh
):
    pass


class VehicleSpeedTooLow(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.vehicleSpeedTooLow
):
    pass


class ThrottlePedalTooHigh(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.throttlePedalTooHigh
):
    pass


class ThrottlePedalTooLow(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.throttlePedalTooLow
):
    pass


class TransmissionRangeNotInNeutral(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.transmissionRangeNotInNeutral,
):
    pass


class TransmissionRangeNotInGear(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.transmissionRangeNotInGear
):
    pass


class BrakeSwitchNotClosed(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.brakeSwitchNotClosed
):
    pass


class ShifterLeverNotInPark(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.shifterLeverNotInPark
):
    pass


class TorqueConverterClutchLocked(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.torqueConverterClutchLocked
):
    pass


class VoltageTooHigh(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.voltageTooHigh
):
    pass


class VoltageTooLow(
    UnexpectedNegativeResponse, response_code=UDSErrorCodes.voltageTooLow
):
    pass


class ResourceTemporarilyNotAvailable(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.resourceTemporarilyNotAvailable,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF0(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF0,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF1(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF1,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF2(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF2,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF3(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF3,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF4(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF4,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF5(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF5,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF6(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF6,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF7(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF7,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF8(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF8,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectF9(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectF9,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectFA(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectFA,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectFB(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectFB,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectFC(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectFC,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectFD(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectFD,
):
    pass


class VehicleManufacturerSpecificConditionsNotCorrectFE(
    UnexpectedNegativeResponse,
    response_code=UDSErrorCodes.vehicleManufacturerSpecificConditionsNotCorrectFE,
):
    pass
